import datetime
import json

import pytest
import pytz
import responses
from six.moves.urllib_parse import urlparse, parse_qs
import trafaret

from datarobot import BacktestSpecification, DatetimePartitioning, \
    DatetimePartitioningSpecification, FeatureSettings, Project
from datarobot import partitioning_methods
from datarobot.errors import DataRobotDeprecationWarning
from tests.test_project import prep_successful_aim_responses


def test_future_proof(datetime_partition_server_data):
    future_response = dict(datetime_partition_server_data, future='newKey')
    DatetimePartitioning.from_server_data(future_response)


def test_construct_duration_string():
    duration = partitioning_methods.construct_duration_string(years=1, months=2, days=3,
                                                              hours=4, minutes=5, seconds=6)
    assert duration == 'P1Y2M3DT4H5M6S'


def test_construct_duration_string_empty():
    assert partitioning_methods.construct_duration_string() == 'P0Y0M0DT0H0M0S'


def test_collect_payload_fails_cleanly_on_type_error(datetime_partition_spec):
    datetime_partition_spec.holdout_start_date = str(datetime_partition_spec.holdout_start_date)
    with pytest.raises(ValueError) as exc_info:
        datetime_partition_spec.collect_payload()
    assert 'expected holdout_start_date to be a datetime.datetime' in str(exc_info.value)


def test_backtest_to_specification(datetime_partition):
    backtest = datetime_partition.backtests[0]
    backtest_spec = backtest.to_specification()
    assert backtest_spec.index == backtest.index
    assert backtest_spec.gap_duration == backtest.gap_duration
    assert backtest_spec.validation_start_date
    assert backtest_spec.validation_duration == backtest.validation_duration


def test_partition_to_specification(datetime_partition):
    part_spec = datetime_partition.to_specification()
    assert part_spec.datetime_partition_column == datetime_partition.datetime_partition_column
    assert (part_spec.autopilot_data_selection_method ==
            datetime_partition.autopilot_data_selection_method)
    assert part_spec.validation_duration == datetime_partition.validation_duration
    assert part_spec.holdout_start_date == datetime_partition.holdout_start_date
    assert part_spec.holdout_duration == datetime_partition.holdout_duration
    assert part_spec.gap_duration == datetime_partition.gap_duration
    assert part_spec.number_of_backtests == datetime_partition.number_of_backtests
    assert len(part_spec.backtests) == len(datetime_partition.backtests)
    for bt_spec in part_spec.backtests:
        assert isinstance(bt_spec, BacktestSpecification)
    assert part_spec.use_time_series == datetime_partition.use_time_series
    assert part_spec.default_to_known_in_advance == datetime_partition.default_to_known_in_advance
    assert (part_spec.feature_derivation_window_start ==
            datetime_partition.feature_derivation_window_start)
    assert (part_spec.feature_derivation_window_end ==
            datetime_partition.feature_derivation_window_end)
    assert part_spec.forecast_window_start == datetime_partition.forecast_window_start
    assert part_spec.forecast_window_end == datetime_partition.forecast_window_end


def test_partition_to_specification_with_time_series(datetime_partition):
    datetime_partition.multiseries_id_columns = ['series_id']
    datetime_partition.differencing_method = 'seasonal'
    datetime_partition.treat_as_exponential = 'never'
    datetime_partition.periodicities = []
    datetime_partition.use_time_series = True
    datetime_partition.feature_settings = [FeatureSettings('special', known_in_advance=True)]
    part_spec = datetime_partition.to_specification()

    assert part_spec.multiseries_id_columns == datetime_partition.multiseries_id_columns
    assert part_spec.differencing_method == datetime_partition.differencing_method
    assert part_spec.treat_as_exponential == datetime_partition.treat_as_exponential
    assert part_spec.periodicities == datetime_partition.periodicities
    assert part_spec.use_time_series == datetime_partition.use_time_series
    assert part_spec.feature_settings == datetime_partition.feature_settings


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve(project_id, project_url, datetime_partition_after_target_server_data):
    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_after_target_server_data)

    partition = DatetimePartitioning.get(project_id)

    assert partition.primary_training_row_count == 100
    assert partition.available_training_row_count == 100
    assert partition.gap_row_count == 100
    assert partition.holdout_row_count == 100
    assert partition.total_row_count == 100

    backtest = partition.backtests[0]
    assert backtest.primary_training_row_count == 100
    assert backtest.available_training_row_count == 100
    assert backtest.gap_row_count == 100
    assert backtest.validation_row_count == 100
    assert backtest.total_row_count == 100


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve_time_series(
    project_id, project_url, datetime_partition_time_series_server_data
):
    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_time_series_server_data)

    partition = DatetimePartitioning.get(project_id)

    fdw_start = datetime_partition_time_series_server_data['featureDerivationWindowStart']
    fdw_end = datetime_partition_time_series_server_data['featureDerivationWindowEnd']
    fw_start = datetime_partition_time_series_server_data['forecastWindowStart']
    fw_end = datetime_partition_time_series_server_data['forecastWindowEnd']
    assert partition.use_time_series
    assert partition.default_to_known_in_advance
    assert partition.feature_derivation_window_start == fdw_start
    assert partition.feature_derivation_window_end == fdw_end
    assert partition.forecast_window_start == fw_start
    assert partition.forecast_window_end == fw_end


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate(project_id, project_url, datetime_partition_server_data,
                  holdout_start_date, datetime_partition_spec):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data)

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)

    # test that we send the right things
    actual_payload = json.loads(responses.calls[0].request.body)
    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in datetime_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': datetime_partition_spec.datetime_partition_column,
        'autopilotDataSelectionMethod': datetime_partition_spec.autopilot_data_selection_method,
        'validationDuration': datetime_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': datetime_partition_spec.holdout_duration,
        'gapDuration': datetime_partition_spec.gap_duration,
        'numberOfBacktests': datetime_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'useTimeSeries': datetime_partition_spec.use_time_series,
        'defaultToKnownInAdvance': datetime_partition_spec.default_to_known_in_advance
    }
    assert actual_payload == expected_payload

    # test that we get the right thing back
    assert partition.project_id == project_id
    assert partition.multiseries_id_columns is None
    assert (partition.datetime_partition_column ==
            datetime_partition_server_data['datetimePartitionColumn'])
    assert partition.use_time_series == datetime_partition_server_data['useTimeSeries']
    assert (partition.default_to_known_in_advance ==
            datetime_partition_server_data['defaultToKnownInAdvance'])
    assert partition.date_format == datetime_partition_server_data['dateFormat']
    assert (partition.autopilot_data_selection_method ==
            datetime_partition_server_data['autopilotDataSelectionMethod'])
    assert partition.validation_duration == datetime_partition_server_data['validationDuration']

    assert isinstance(partition.available_training_start_date, datetime.datetime)
    assert (partition.available_training_duration ==
            datetime_partition_server_data['availableTrainingDuration'])
    assert isinstance(partition.available_training_end_date, datetime.datetime)

    assert isinstance(partition.primary_training_start_date, datetime.datetime)
    assert (partition.primary_training_duration ==
            datetime_partition_server_data['primaryTrainingDuration'])
    assert isinstance(partition.primary_training_end_date, datetime.datetime)

    assert isinstance(partition.gap_start_date, datetime.datetime)
    assert partition.gap_duration == datetime_partition_server_data['gapDuration']
    assert isinstance(partition.gap_end_date, datetime.datetime)

    assert partition.holdout_start_date == holdout_start_date
    assert partition.holdout_duration == datetime_partition_server_data['holdoutDuration']
    assert isinstance(partition.holdout_end_date, datetime.datetime)

    assert partition.number_of_backtests == datetime_partition_server_data['numberOfBacktests']
    [backtest] = partition.backtests
    [backtest_data] = datetime_partition_server_data['backtests']
    assert backtest.index == backtest_data['index']
    assert isinstance(backtest.available_training_start_date, datetime.datetime)
    assert backtest.available_training_duration == backtest_data['availableTrainingDuration']
    assert isinstance(backtest.available_training_end_date, datetime.datetime)
    assert isinstance(backtest.primary_training_start_date, datetime.datetime)
    assert backtest.primary_training_duration == backtest_data['primaryTrainingDuration']
    assert isinstance(backtest.primary_training_end_date, datetime.datetime)
    assert isinstance(backtest.gap_start_date, datetime.datetime)
    assert backtest.gap_duration == backtest_data['gapDuration']
    assert isinstance(backtest.gap_end_date, datetime.datetime)
    assert isinstance(backtest.validation_start_date, datetime.datetime)
    assert backtest.validation_duration == backtest_data['validationDuration']
    assert isinstance(backtest.validation_end_date, datetime.datetime)


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate_with_multiseries(project_id, project_url, datetime_partition_spec,
                                   datetime_partition_server_data):
    datetime_partition_spec.multiseries_id_columns = ['series_id']
    datetime_partition_server_data['multiseriesIdColumns'] = ['series_id']
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, datetime_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': datetime_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    date_feat_json = {
        'projectId': project_id, 'name': datetime_partition_spec.datetime_partition_column,
        'id': 2, 'timeSeriesEligible': False, 'featureType': 'date',
        'timeSeriesEligibilityReason': 'notUnique', 'lowInformation': False, 'uniqueCount': 500
    }
    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          datetime_partition_spec.datetime_partition_column),
                  json=date_feat_json)
    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)
    assert partition.multiseries_id_columns == ['series_id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate_time_series(project_id, project_url, datetime_partition_time_series_server_data,
                              time_series_partition_spec, holdout_start_date):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_time_series_server_data)

    partition = DatetimePartitioning.generate(project_id, time_series_partition_spec)

    # test that we send the right things
    actual_payload = json.loads(responses.calls[0].request.body)
    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end
    }
    assert actual_payload == expected_payload

    assert partition.use_time_series == time_series_partition_spec.use_time_series
    assert (partition.default_to_known_in_advance ==
            time_series_partition_spec.default_to_known_in_advance)
    assert (partition.feature_derivation_window_start ==
            time_series_partition_spec.feature_derivation_window_start)
    assert (partition.feature_derivation_window_end ==
            time_series_partition_spec.feature_derivation_window_end)
    assert partition.forecast_window_start == time_series_partition_spec.forecast_window_start
    assert partition.forecast_window_end == time_series_partition_spec.forecast_window_end


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate_feature_settings(
        project_id, project_url, datetime_partition_server_data_feature_settings,
        holdout_start_date, datetime_partition_spec_feature_settings
):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data_feature_settings)

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec_feature_settings)

    # test that we send the right things
    actual_payload = json.loads(responses.calls[0].request.body)
    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
    } for bt in datetime_partition_spec_feature_settings.backtests]
    expected_feature_settings = [{
        'featureName': fs.feature_name,
        'knownInAdvance': fs.known_in_advance
    } for fs in datetime_partition_spec_feature_settings.feature_settings]
    expected_payload = {
        'datetimePartitionColumn':
            datetime_partition_spec_feature_settings.datetime_partition_column,
        'autopilotDataSelectionMethod':
            datetime_partition_spec_feature_settings.autopilot_data_selection_method,
        'validationDuration': datetime_partition_spec_feature_settings.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': datetime_partition_spec_feature_settings.holdout_duration,
        'gapDuration': datetime_partition_spec_feature_settings.gap_duration,
        'numberOfBacktests': datetime_partition_spec_feature_settings.number_of_backtests,
        'backtests': expected_backtests,
        'useTimeSeries': datetime_partition_spec_feature_settings.use_time_series,
        'defaultToKnownInAdvance':
            datetime_partition_spec_feature_settings.default_to_known_in_advance,
        'featureSettings': expected_feature_settings
    }
    assert actual_payload == expected_payload
    assert partition.feature_settings == datetime_partition_spec_feature_settings.feature_settings


@pytest.mark.parametrize('known_in_advance', [None, 'True', 1, {'true': True}, [True]])
def test_create_feature_settings_bad_known_in_advance(known_in_advance):
    with pytest.raises(trafaret.DataError):
        FeatureSettings('feature_name', known_in_advance=known_in_advance)


@pytest.mark.parametrize('feature_name', [None, False, 1, {'true': True}, [True]])
def test_create_feature_settings_bad_feature_name(feature_name):
    with pytest.raises(trafaret.DataError):
        FeatureSettings(feature_name, 'known_in_advance')


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate_keep_extra_time_series_controls_spec(
    project_id, project_url,
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=server_data)

    partitioning = DatetimePartitioning.generate(project_id, spec_fixture)
    partitioning_spec = partitioning.to_specification()

    assert partitioning_spec.periodicities == spec_fixture.periodicities
    assert partitioning_spec.treat_as_exponential == spec_fixture.treat_as_exponential
    assert partitioning_spec.differencing_method == spec_fixture.differencing_method


@responses.activate
@pytest.mark.usefixtures('client')
def test_generate_no_peridiocities_on_server_spec(
    project_id, project_url,
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    server_data['periodicities'] = None
    spec_fixture.periodicities = None
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=server_data)

    partitioning = DatetimePartitioning.generate(project_id, spec_fixture)
    partitioning_spec = partitioning.to_specification()

    assert partitioning_spec.periodicities == spec_fixture.periodicities
    assert partitioning_spec.treat_as_exponential == spec_fixture.treat_as_exponential
    assert partitioning_spec.differencing_method == spec_fixture.differencing_method


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_project_with_periodicity(
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project('p-id').set_target(
        'SalePrice',
        metric='RMSE',
        partitioning_method=spec_fixture)

    request_message = json.loads(responses.calls[0].request.body)
    assert request_message['periodicities'] == server_data['periodicities']


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_project_with_treat_as_exponential(
    datetime_partition_spec_extra_time_series_controls
):
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project('p-id').set_target(
        'SalePrice',
        metric='RMSE',
        partitioning_method=spec_fixture)

    request_message = json.loads(responses.calls[0].request.body)
    assert request_message['treatAsExponential'] == spec_fixture.treat_as_exponential


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_project_with_differencing_method(
        datetime_partition_spec_extra_time_series_controls
):
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project('p-id').set_target(
        'SalePrice',
        metric='RMSE',
        partitioning_method=spec_fixture)

    request_message = json.loads(responses.calls[0].request.body)
    assert request_message['differencingMethod'] == spec_fixture.differencing_method


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve_otp_without_holdout(project_id, project_url,
                                      datetime_partition_without_holdout_server_data):
    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_without_holdout_server_data)

    partition = DatetimePartitioning.get(project_id)

    assert partition.primary_training_start_date is None
    assert partition.primary_training_end_date is None
    assert partition.primary_training_row_count == 0
    assert partition.primary_training_duration == 'P0Y0M0D'

    assert partition.available_training_start_date == datetime.datetime(2013, 11, 20, 22, 38, 47,
                                                                        tzinfo=pytz.UTC)
    assert partition.available_training_end_date == datetime.datetime(2014, 1, 12, 22, 44, 23,
                                                                      tzinfo=pytz.UTC)
    assert partition.available_training_duration == 'P0Y0M53DT0H5M36S'
    assert partition.available_training_row_count == 4276

    assert partition.gap_start_date is None
    assert partition.gap_end_date is None
    assert partition.gap_row_count == 0
    assert partition.gap_duration == 'P0Y0M0D'

    assert partition.holdout_start_date is None
    assert partition.holdout_end_date is None
    assert partition.holdout_row_count == 0
    assert partition.holdout_duration == 'P0Y0M0D'

    assert partition.total_row_count == 4276

    backtest = partition.backtests[0]
    assert backtest.primary_training_row_count == 4168
    assert backtest.available_training_row_count == 4168
    assert backtest.gap_row_count == 0
    assert backtest.validation_row_count == 108
    assert backtest.total_row_count == 4276


@responses.activate
@pytest.mark.usefixtures('client')
def test_retrieve_feature_settings(
        project_id, project_url, datetime_partition_after_target_server_data_feature_settings):

    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_after_target_server_data_feature_settings)

    partition = DatetimePartitioning.get(project_id)

    fs_list = datetime_partition_after_target_server_data_feature_settings['featureSettings']
    assert partition.feature_settings == [
        FeatureSettings(fs['featureName'], fs['knownInAdvance']) for fs in fs_list]


@responses.activate
@pytest.mark.usefixtures('client')
def test_from_minimum_specification(project_id, project_url, datetime_partition_server_data):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_server_data)

    datetime_column = datetime_partition_server_data['datetimePartitionColumn']
    min_spec = DatetimePartitioningSpecification(datetime_column)
    DatetimePartitioning.generate(project_id, min_spec)

    expected_payload = {'datetimePartitionColumn': datetime_column,
                        'useTimeSeries': False, 'defaultToKnownInAdvance': False}
    actual_payload = json.loads(responses.calls[0].request.body)
    assert actual_payload == expected_payload


def test_to_dataframe(datetime_partition):
    df = datetime_partition.to_dataframe()
    expected_shape = (4*(1+datetime_partition.number_of_backtests), 3)
    assert df.shape == expected_shape


@responses.activate
@pytest.mark.usefixtures('client')
def test_disabled_holdout(project_id, project_url, datetime_partition_without_holdout_server_data):
    responses.add(responses.POST, '{}datetimePartitioning/'.format(project_url),
                  json=datetime_partition_without_holdout_server_data)

    datetime_column = datetime_partition_without_holdout_server_data['datetimePartitionColumn']
    spec = DatetimePartitioningSpecification(datetime_column, disable_holdout=True)
    DatetimePartitioning.generate(project_id, spec)

    expected_payload = {
        'datetimePartitionColumn': datetime_column,
        'disableHoldout': True,
        'useTimeSeries': False,
        'defaultToKnownInAdvance': False
    }
    actual_payload = json.loads(responses.calls[0].request.body)
    assert actual_payload == expected_payload


# APriori backward compatibility
@responses.activate
@pytest.mark.usefixtures('client', 'known_warning')
def test_a_priori_backward_compatibility_supported(
        project_id, project_url, datetime_partition_after_target_server_data_feature_settings):

    resp = datetime_partition_after_target_server_data_feature_settings
    resp['defaultToAPriori'] = resp.pop('defaultToKnownInAdvance')
    resp['featureSettings'] = [
        {'aPriori': r['knownInAdvance'], 'featureName': r['featureName']}
        for r in resp['featureSettings']
    ]
    responses.add(responses.GET, '{}datetimePartitioning/'.format(project_url), json=resp)

    partition = DatetimePartitioning.get(project_id)

    fs_list = datetime_partition_after_target_server_data_feature_settings['featureSettings']
    # Test response with aPriori works
    assert partition.feature_settings == [
        FeatureSettings(fs['featureName'], fs['aPriori']) for fs in fs_list]
    # Test a_priori works
    assert partition.feature_settings[0].a_priori
    # Test default_to_a_priori works
    assert partition.default_to_a_priori == resp['defaultToAPriori']


@pytest.mark.parametrize('a_priori', ['True', 1, {'true': True}, [True]])
def test_create_feature_settings_bad_a_priori(a_priori):
    with pytest.raises(DataRobotDeprecationWarning):
        FeatureSettings('feature_name', a_priori=a_priori)


@responses.activate
@pytest.mark.usefixtures('client')
def test_time_series_feature_log(project_id, project_url):
    expected_resp = {
        'count': 0,
        'previous': None,
        'totalLogLines': 78,
        'featureLog': '',
        'next': None
    }
    responses.add(responses.GET, '{}timeSeriesFeatureLog/'.format(project_url), json=expected_resp)
    resp = DatetimePartitioning.feature_log_list(project_id, limit=10, offset=20)
    assert resp == expected_resp

    req_url = responses.calls[0].request.url
    assert {'limit': ['10'], 'offset': ['20']} == parse_qs(urlparse(req_url).query)


@responses.activate
@pytest.mark.usefixtures('client')
def test_time_series_feature_log_retrieve(project_id, project_url):
    log_data = 'useful info'
    responses.add(responses.GET, '{}timeSeriesFeatureLog/file/'.format(project_url), body=log_data)
    retrieved_data = DatetimePartitioning.feature_log_retrieve(project_id)
    assert retrieved_data == log_data



import json

import datarobot
import pytest
import responses

from datarobot import Project, Blueprint
from datarobot import enums


@responses.activate
@pytest.mark.usefixtures('client')
def test_async_flow(project, project_id, model_job_json, model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=202,
        body='',
        adding_headers={'Location': 'https://host_name.com/projects/p-id/modelJobs/12/'}
    )
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/modelJobs/12/'.format(project_id),
        content_type='application/json',
        status=200,
        body=model_job_json
    )

    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    retval = project.train(blueprint_id)
    assert retval == '12'


def test_print_project_from_id():
    project = Project('some-project-id')
    print(project)  # checks that method relevant to print works on this (failed in one release)


@responses.activate
@pytest.mark.usefixtures('client')
def test_all_defaults(project,
                      model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://host_name.com/projects/p-id/modelJobs/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    model_id = project.train(blueprint_id)
    assert responses.calls[0].request.url == model_collection_url
    payload = json.loads(responses.calls[0].request.body)
    assert 'samplePct' not in payload
    assert 'trainingRowCount' not in payload
    assert 'id' == model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_no_defaults(project,
                     model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://host_name.com/projects/p-id/modelJobs/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    dataset_id = '5223deadbeefdeadbeef0101'
    source_project_id = '5223deadbeefdeadbeef1234'
    scoring_type = datarobot.SCORING_TYPE.cross_validation
    sample_pct = 44

    project.train(blueprint_id,
                  featurelist_id=dataset_id,
                  source_project_id=source_project_id,
                  sample_pct=sample_pct,
                  scoring_type=scoring_type)
    assert responses.calls[0].request.url == model_collection_url
    request = json.loads(responses.calls[0].request.body)
    assert request['blueprintId'] == blueprint_id
    assert request['sourceProjectId'] == source_project_id
    assert request['featurelistId'] == dataset_id
    assert request['samplePct'] == sample_pct
    assert request['scoringType'] == scoring_type
    assert 'trainingRowCount' not in request


def test_train_cant_use_both(project):
    blueprint_id = 'a-blueprint'
    with pytest.raises(ValueError):
        project.train(blueprint_id, sample_pct=40, training_row_count=100)


@responses.activate
@pytest.mark.usefixtures('client')
def test_train_by_rowcount_with_bp_id(project, model_collection_url, project_url):
    responses.add(responses.POST, model_collection_url, content_type='application/json',
                  status=201, body='',
                  adding_headers={'Location': '{}/modelJobs/5/'.format(project_url)})
    blueprint_id = 'a-blueprint-id-hash'
    training_row_count = 50
    project.train(blueprint_id, training_row_count=training_row_count)
    assert responses.calls[0].request.url == model_collection_url
    payload = json.loads(responses.calls[0].request.body)
    assert payload['blueprintId'] == blueprint_id
    assert payload['trainingRowCount'] == training_row_count
    assert 'samplePct' not in payload


@responses.activate
@pytest.mark.usefixtures('client')
def test_train_by_rowcount_with_trainable(project, model_collection_url, project_url):
    responses.add(responses.POST, model_collection_url, content_type='application/json',
                  status=201, body='',
                  adding_headers={'Location': '{}/modelJobs/5/'.format(project_url)})
    blueprint_id = 'a-blueprint-hash'
    trainable = Blueprint.from_data({'id': blueprint_id, 'project_id': project.id,
                                     'model_type': 'A fake model',
                                     'processes': ['Things', 'Stuff']})
    training_row_count = 50

    project.train(trainable, training_row_count=training_row_count)
    payload = json.loads(responses.calls[0].request.body)
    assert payload['blueprintId'] == blueprint_id
    assert payload['trainingRowCount'] == training_row_count
    assert 'samplePct' not in payload


@responses.activate
@pytest.mark.usefixtures('client')
def test_with_trainable_object_ignores_any_source_project_id(project,
                                                             model_collection_url):

    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://pam/api/v2/projects/p-id/models/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    source_project_id = '5223deadbeefdeadbeef1234'

    data = dict(id=blueprint_id,
                project_id=source_project_id,
                model_type='Pretend Model',
                processes=['Cowboys', 'Aliens'])

    blueprint = Blueprint.from_data(data)

    project.train(blueprint,
                  source_project_id='should-be-ignored')
    assert responses.calls[0].request.url == model_collection_url
    request = json.loads(responses.calls[0].request.body)
    assert request['blueprintId'] == blueprint_id
    assert request['sourceProjectId'] == source_project_id


MONO_INC_FL_ID = '5ae1afb6962d74625f940991'
MONO_DEC_FL_ID = '5ae1afb6962d74625f940992'
BLUEPRINT_ID = '5a997601962d742269a38dd9'


@pytest.mark.parametrize('blueprint,featurelists', [
    (Blueprint(id=BLUEPRINT_ID), (MONO_INC_FL_ID, MONO_DEC_FL_ID)),
    (BLUEPRINT_ID, (MONO_INC_FL_ID, MONO_DEC_FL_ID)),
    (BLUEPRINT_ID, (None, None)),
    (BLUEPRINT_ID, (MONO_INC_FL_ID, None)),
    (BLUEPRINT_ID, (None, MONO_DEC_FL_ID)),
    (BLUEPRINT_ID, (enums.MONOTONICITY_FEATURELIST_DEFAULT, MONO_DEC_FL_ID)),
])
@responses.activate
@pytest.mark.usefixtures('client')
def test_monotonicity_blueprint(blueprint, project, featurelists):
    monotonic_increasing_featurelist_id, monotonic_decreasing_featurelist_id = featurelists

    responses.add(
        responses.POST,
        'https://host_name.com/projects/{}/models/'.format(project.id),
        content_type='application/json',
        status=201,
        body='',
        adding_headers={
            'Location': 'https://host_name.com/projects/p-id/5a9975c9962d746034393e42/modelJobs/1/'
        }
    )

    project.train(
        blueprint,
        monotonic_increasing_featurelist_id=monotonic_increasing_featurelist_id,
        monotonic_decreasing_featurelist_id=monotonic_decreasing_featurelist_id,
    )
    payload = json.loads(responses.calls[0].request.body)
    if monotonic_increasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert 'monotonicIncreasingFeaturelistId' not in payload
    else:
        assert payload['monotonicIncreasingFeaturelistId'] == monotonic_increasing_featurelist_id

    if monotonic_decreasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert 'monotonicDecreasingFeaturelistId' not in payload
    else:
        assert payload['monotonicDecreasingFeaturelistId'] == monotonic_decreasing_featurelist_id



from six.moves.urllib_parse import urlencode
import trafaret as t
from datarobot import enums
from ..utils import from_api, raw_prediction_response_to_dataframe, encode_utf8_if_py2
from .api_object import APIObject

_base_path = 'projects/{}/predictions/'
_get_path = _base_path + '{}/'


class Predictions(APIObject):
    """
    Represents predictions metadata and provides access to prediction results.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model
    prediction_id : str
        id of generated predictions

    Examples
    --------

    List all predictions for a project

    .. code-block:: python

        import datarobot as dr

        # Fetch all predictions for a project
        all_predictions = dr.Predictions.list(project_id)

        # Inspect all calculated predictions
        for predictions in all_predictions:
            print(predictions)  # repr includes project_id, model_id, and dataset_id

    Retrieve predictions by id

    .. code-block:: python

        import datarobot as dr

        # Getting predictions by id
        predictions = dr.Predictions.get(project_id, prediction_id)

        # Dump actual predictions
        df = predictions.get_all_as_dataframe()
        print(df)
    """

    def __init__(self, project_id, prediction_id, model_id=None, dataset_id=None):
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.prediction_id = prediction_id
        self.path = _get_path.format(self.project_id, self.prediction_id)

    @classmethod
    def _build_list_path(cls, project_id, model_id=None, dataset_id=None):
        args = {}
        if model_id:
            args['modelId'] = model_id
        if dataset_id:
            args['datasetId'] = dataset_id

        path = _base_path.format(project_id)
        if args:
            path = '{}?{}'.format(path, urlencode(args))

        return path

    @classmethod
    def list(cls, project_id, model_id=None, dataset_id=None):
        """
        Fetch all the computed predictions metadata for a project.

        Parameters
        ----------
        project_id : str
            id of the project
        model_id : str, optional
            if specified, only predictions metadata for this model will be retrieved
        dataset_id : str, optional
            if specified, only predictions metadata for this dataset will be retrieved

        Returns
        -------
        A list of :py:class:`Predictions <datarobot.models.Predictions>` objects
        """
        _trafaret = t.Dict({
            t.Key('data'): t.List(t.Dict({
                t.Key('id'): t.String(),
                t.Key('url'): t.String(),
                t.Key('model_id'): t.String(),
                t.Key('dataset_id'): t.String(),
            }).ignore_extra('*')),
        }).ignore_extra('*')

        path = cls._build_list_path(project_id, model_id=model_id, dataset_id=dataset_id)
        converted = from_api(cls._server_data(path))
        validated = _trafaret.check(converted)['data']
        return [
            cls(project_id,
                prediction_id=item['id'],
                model_id=item['model_id'],
                dataset_id=item['dataset_id'])
            for item in validated]

    @classmethod
    def get(cls, project_id, prediction_id):
        """
        Retrieve the specific predictions metadata

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        prediction_id : str
            id of the prediction set

        Returns
        -------
        :py:class:`Predictions <datarobot.models.Predictions>` object representing specified
        predictions
        """
        return cls(project_id, prediction_id)

    def get_all_as_dataframe(self, class_prefix=enums.PREDICTION_PREFIX.DEFAULT):
        """
        Retrieve all prediction rows and return them as a pandas.DataFrame.

        Parameters
        ----------
        class_prefix : str, optional
            The prefix to append to labels in the final dataframe. Default is ``class_``
            (e.g., apple -> class_apple)

        Returns
        -------
        dataframe: pandas.DataFrame
        """
        data = self._server_data(self.path)
        return raw_prediction_response_to_dataframe(data, class_prefix)

    def download_to_csv(self, filename, encoding='utf-8'):
        """
        Save prediction rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save prediction rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to
            'utf-8'
        """
        df = self.get_all_as_dataframe()
        df.to_csv(
            path_or_buf=filename,
            header=True,
            index=False,
            encoding=encoding,
        )

    def __repr__(self):
        template = u'{}(prediction_id={!r}, project_id={!r}, model_id={!r}, dataset_id={!r})'
        return encode_utf8_if_py2(template.format(
            type(self).__name__,
            self.prediction_id,
            self.project_id,
            self.model_id,
            self.dataset_id))



import mock
import json

import pytest
import responses


@pytest.yield_fixture
def mock_async_time():
    with mock.patch('datarobot.utils.waiters.time') as mock_time:
        yield mock_time


@pytest.fixture
def async_url():
    return 'https://host_name.com/status/status-id/'


@pytest.fixture
def model_collection_url(project_id):
    return 'https://host_name.com/projects/{}/models/'.format(project_id)


@pytest.fixture
def model_job_data(project_id, model_id):
    return {'status': 'inprogress',
            'samplePct': 64.0,
            'processes': ['One-Hot Encoding',
                          'Missing Values Imputed',
                          'Standardize',
                          'Linear Regression'],
            'modelType': 'Linear Regression',
            'featurelistId': '55666d05100d2b01a1104dae',
            'blueprintId': '3bb4665320be633b30a9601b3e73284d',
            'projectId': project_id,
            'id': 12,
            'modelId': model_id}


@pytest.fixture
def model_job_json(model_job_data):
    return json.dumps(model_job_data)


@pytest.fixture
def project_creation_responses(async_url,
                               project_collection_url,
                               project_url,
                               project_without_target_json):
    responses.add(responses.POST,
                  project_collection_url,
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': async_url}
                  )
    responses.add(responses.GET,
                  async_url,
                  status=303,
                  body='',
                  content_type='application/json',
                  adding_headers={
                      'Location': project_url}
                  )
    responses.add(responses.GET,
                  project_url,
                  status=200,
                  body=project_without_target_json,
                  content_type='application/json')


@pytest.fixture
def project_aim_responses(project_aim_url,
                          async_url,
                          project_url,
                          project_with_target_json):
    responses.add(
        responses.PATCH,
        project_aim_url,
        body='',
        status=202,
        adding_headers={'Location': async_url},
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        async_url,
        body='',
        status=303,
        adding_headers={'Location': project_url},
        content_type='application/json')

    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type='application/json',
        body=project_with_target_json
    )



# coding=utf-8
import json

import pytest
import responses
from numpy import isclose

from datarobot import Model
from datarobot.models.word_cloud import WordCloud


@pytest.fixture
def word_cloud_data():
    return {'ngrams': [
        {
            'ngram': u'エピソード',
            'coefficient': 0.2074072186419962,
            'isStopword': False,
            'count': 10,
            'frequency': 0.1449275362318841
        },
        {
            'ngram': u'外',
            'coefficient': 0.3669030452411179,
            'isStopword': True,
            'count': 8,
            'frequency': 0.1159420289855072
        },
        {
            'ngram': u'閉鎖',
            'coefficient': -0.1343662758815528,
            'isStopword': False,
            'count': 5,
            'frequency': 0.07246376811594203
        }]}


def test_instantiation(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)

    assert len(word_cloud_data['ngrams']) == len(wc.ngrams)
    for i in range(len(wc.ngrams)):
        assert wc.ngrams[i]['ngram'] == word_cloud_data['ngrams'][i]['ngram']
        assert wc.ngrams[i]['is_stopword'] == word_cloud_data['ngrams'][i]['isStopword']
        assert wc.ngrams[i]['count'] == word_cloud_data['ngrams'][i]['count']
        assert isclose(wc.ngrams[i]['coefficient'], word_cloud_data['ngrams'][i]['coefficient'])
        assert isclose(wc.ngrams[i]['frequency'], word_cloud_data['ngrams'][i]['frequency'])


def test_future_proof(word_cloud_data):
    data_with_future_keys = dict(word_cloud_data, new_key='new_value')
    data_with_future_keys['ngrams'][0]['some_additional_data'] = 'we never expected this'
    WordCloud.from_server_data(data_with_future_keys)


def test_most_frequent(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)
    most_frequent_list = wc.most_frequent(1)
    assert len(most_frequent_list) == 1
    most_frequent = most_frequent_list[0]
    assert all(most_frequent['frequency'] >= wc_ngram['frequency']
               for wc_ngram in wc.ngrams)


def test_most_important(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)
    most_important_list = wc.most_important(1)
    assert len(most_important_list) == 1
    most_important = most_important_list[0]
    assert all(abs(most_important['coefficient']) >= abs(wc_ngram['coefficient'])
               for wc_ngram in wc.ngrams)


@pytest.fixture
def word_cloud_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/wordCloud/'.format(
        project_id,
        model_id
    )


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_word_cloud(word_cloud_data,
                        word_cloud_url,
                        project_id, model_id):
    responses.add(
        responses.GET,
        word_cloud_url,
        status=200,
        content_type='application/json',
        body=json.dumps(word_cloud_data)
    )
    model = Model(id=model_id, project_id=project_id)
    wc = model.get_word_cloud()

    assert len(word_cloud_data['ngrams']) == len(wc.ngrams)
    for i in range(len(wc.ngrams)):
        assert wc.ngrams[i]['ngram'] == word_cloud_data['ngrams'][i]['ngram']
        assert wc.ngrams[i]['is_stopword'] == word_cloud_data['ngrams'][i]['isStopword']
        assert wc.ngrams[i]['count'] == word_cloud_data['ngrams'][i]['count']
        assert isclose(wc.ngrams[i]['coefficient'], word_cloud_data['ngrams'][i]['coefficient'])
        assert isclose(wc.ngrams[i]['frequency'], word_cloud_data['ngrams'][i]['frequency'])



"""
  AbstractJob
  ^        ^
  |        |
  +        +
Job       AbstractSpecificJob
            ^             ^
            |             |
            +             +
        ModelJob       PredictJob


Subclasses of AbstractJob can override _converter_extra and _extra_fields to indicate extra object
attributes which come in the job data. They must implement the methods which raise
NotImplementedError

The Job class represents jobs of any type (whether or not we have classes for more specific job
types) with only the data and methods that work for jobs of any type.

We also have classes representing specific jobs. Some of the functionality for these jobs is shared
and hence is implemented in AbstractSpecificJob.

AbstractSpecificJob and AbstractJob should be non-public and may change. Users should only rely
on the concrete classes.
"""
import trafaret as t

from datarobot.utils.waiters import wait_for_async_resolution
from .model import Model, PrimeModel
from .rating_table import RatingTable
from .prediction_explanations import PredictionExplanationsInitialization, PredictionExplanations
from .prime_file import PrimeFile
from .reason_codes import ReasonCodesInitialization, ReasonCodes
from .ruleset import Ruleset
from .training_predictions import TrainingPredictions
from ..client import get_client, staticproperty
from ..enums import JOB_TYPE, QUEUE_STATUS, DEFAULT_MAX_WAIT, PREDICTION_PREFIX
from ..utils import from_api, raw_prediction_response_to_dataframe
from .. import errors


class AbstractJob(object):
    """ Generic representation of a job in the queue
    """
    # Subclasses can override this:
    _converter_extra = t.Dict()
    _extra_fields = frozenset()

    # Subclasses should not override any of this:
    _client = staticproperty(get_client)
    _converter_common = t.Dict({
        t.Key('id', optional=True): t.Int,
        t.Key('status', optional=True): t.Enum(
            QUEUE_STATUS.ABORTED,
            QUEUE_STATUS.COMPLETED,
            QUEUE_STATUS.ERROR,
            QUEUE_STATUS.INPROGRESS,
            QUEUE_STATUS.QUEUE),
        t.Key('project_id', optional=True): t.String,
        t.Key('is_blocked'): t.Bool,
    })

    def __init__(self, data, completed_resource_url=None):
        # Importing here to dodge circular dependency
        from . import Project

        if not isinstance(data, dict):
            raise ValueError('job data must be a dict')
        self._completed_resource_url = completed_resource_url
        data = from_api(data)

        converter = (self._converter_common + self._converter_extra).allow_extra('*')
        self._safe_data = converter.check(data)
        self.job_type = self._get_job_type(self._safe_data)
        self.status = self._safe_data.get('status')
        self.id = self._safe_data.get('id')
        self.project = Project(self._safe_data.get('project_id'))
        self.project_id = self._safe_data.get('project_id')
        self.is_blocked = self._safe_data.get('is_blocked')
        for k in self._extra_fields:
            v = self._safe_data.get(k)
            setattr(self, k, v)

    @classmethod
    def _get_job_type(cls, safe_data):
        raise NotImplementedError

    def __repr__(self):
        return 'Job({}, status={})'.format(self.job_type, self.status)

    @classmethod
    def _data_and_completed_url_for_job(cls, url):
        response = cls._client.get(url, allow_redirects=False)

        if response.status_code in (200, 303):
            data = response.json()
            completed_url = response.headers['Location'] if response.status_code == 303 else None
            return data, completed_url
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))

    def refresh(self):
        """
        Update this object with the latest job data from the server.
        """
        data, completed_url = self._data_and_completed_url_for_job(self._this_job_path())
        self.__init__(data, completed_resource_url=completed_url)

    def get_result(self):
        """
        Returns
        -------
        result : object
            Return type depends on the job type:
                - for model jobs, a Model is returned
                - for predict jobs, a pandas.DataFrame (with predictions) is returned
                - for featureImpact jobs, a list of dicts (see
                  :meth:`Model.get_feature_impact <datarobot.models.Model.get_feature_impact>` for
                  more detail)
                - for primeRulesets jobs, a list of Rulesets
                - for primeModel jobs, a PrimeModel
                - for primeDownloadValidation jobs, a PrimeFile
                - for reasonCodesInitialization jobs, a ReasonCodesInitialization
                - for reasonCodes jobs, a ReasonCodes
                - for predictionExplanationInitialization jobs, a
                  PredictionExplanationsInitialization
                - for predictionExplanations jobs, a PredictionExplanations

        Raises
        ------
        JobNotFinished
            If the job is not finished, the result is not available.
        AsyncProcessUnsuccessfulError
            If the job errored or was aborted
        """
        self.refresh()
        if self.status in [QUEUE_STATUS.ERROR, QUEUE_STATUS.ABORTED]:
            raise errors.AsyncProcessUnsuccessfulError
        if not self._completed_resource_url:
            raise errors.JobNotFinished
        completed_resource_path = self._client.strip_endpoint(self._completed_resource_url)
        return self._make_result_from_location(completed_resource_path)

    def _make_result_from_json(self, server_data):
        if self.job_type == JOB_TYPE.MODEL:
            return Model.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PREDICT:
            return raw_prediction_response_to_dataframe(server_data, PREDICTION_PREFIX.DEFAULT)
        elif self.job_type == JOB_TYPE.FEATURE_IMPACT:
            return server_data['featureImpacts']
        elif self.job_type == JOB_TYPE.PRIME_RULESETS:
            return [Ruleset.from_server_data(ruleset_data) for ruleset_data in server_data]
        elif self.job_type == JOB_TYPE.PRIME_MODEL:
            return PrimeModel.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PRIME_VALIDATION:
            return PrimeFile.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.REASON_CODES_INITIALIZATION:
            return ReasonCodesInitialization.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PREDICTION_EXPLANATIONS_INITIALIZATION:
            return PredictionExplanationsInitialization.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.REASON_CODES:
            return ReasonCodes.from_location(server_data['reasonCodesRecordLocation'])
        elif self.job_type == JOB_TYPE.PREDICTION_EXPLANATIONS:
            return PredictionExplanations.from_location(
                server_data['predictionExplanationsRecordLocation'])
        elif self.job_type == JOB_TYPE.RATING_TABLE_VALIDATION:
            return RatingTable.from_server_data(server_data)
        else:
            raise ValueError("Unrecognized job type {}.".format(self.job_type))

    def _make_result_from_location(self, location):
        if self.job_type == JOB_TYPE.TRAINING_PREDICTIONS:
            return TrainingPredictions.from_location(location)

        server_data = self._client.get(location).json()
        return self._make_result_from_json(server_data)

    def wait_for_completion(self, max_wait=DEFAULT_MAX_WAIT):
        """
        Waits for job to complete.

        Parameters
        ----------
        max_wait : int, optional
            How long to wait for the job to finish.
        """
        try:
            wait_for_async_resolution(
                self._client,
                self._this_job_path(),
                max_wait=max_wait
            )
        finally:
            # We are gonna try to update the job data, that's OK if it fails too (rare cases)
            self.refresh()

    def get_result_when_complete(self, max_wait=DEFAULT_MAX_WAIT):
        """
        Parameters
        ----------
        max_wait : int, optional
            How long to wait for the job to finish.

        Returns
        -------
        result: object
            Return type is the same as would be returned by `Job.get_result`.

        Raises
        ------
        AsyncTimeoutError
            If the job does not finish in time
        AsyncProcessUnsuccessfulError
            If the job errored or was aborted
        """
        if self.job_type == JOB_TYPE.MODEL_EXPORT:
            # checking this here instead of in _make_result to avoid waiting for the response
            raise ValueError("Can't return the result for a model export job. Use "
                             "Job.wait_for_completion to wait for the job to complete and "
                             "Model.download_export to download the finished export.")

        self.wait_for_completion(max_wait=max_wait)
        return self._make_result_from_location(self._completed_resource_url)

    def _this_job_path(self):
        return self._job_path(self.project.id, self.id)

    @classmethod
    def _job_path(cls, project_id, job_id):
        raise NotImplementedError

    @classmethod
    def get(cls, project_id, job_id):
        # Note: For 3.0 (when the behavior of all job types' `get` methods can be made consistent),
        #       the implementation can move here.
        raise NotImplementedError

    def cancel(self):
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._client.delete(self._this_job_path())


class Job(AbstractJob):

    """ Tracks asynchronous work being done within a project

    Attributes
    ----------
    id : int
        the id of the job
    project_id : str
        the id of the project the job belongs to
    status : str
        the status of the job - will be one of ``datarobot.enums.QUEUE_STATUS``
    job_type : str
        what kind of work the job is doing - will be one of ``datarobot.enums.JOB_TYPE``
    is_blocked : bool
        if true, the job is blocked (cannot be executed) until its dependencies are resolved
    """

    _converter_extra = t.Dict({
        t.Key('job_type', optional=True) >> 'job_type': t.String,
        t.Key('url') >> 'url': t.String
    })

    def __init__(self, data, completed_resource_url=None):
        super(Job, self).__init__(data, completed_resource_url=completed_resource_url)
        self._job_details_path = self._client.strip_endpoint(self._safe_data['url'])

    @classmethod
    def _get_job_type(cls, safe_data):
        # For generic jobs, job_type comes from the data.
        return safe_data.get('job_type')

    @classmethod
    def _job_path(cls, project_id, job_id):
        # Path where you can get jobs of this kind.
        return 'projects/{}/jobs/{}/'.format(project_id, job_id)

    @classmethod
    def get(cls, project_id, job_id):
        """
        Fetches one job.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job id

        Returns
        -------
        job : Job
            The job

        Raises
        ------
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(data, completed_resource_url=completed_url)


class AbstractSpecificJob(AbstractJob):
    @classmethod
    def _job_type(cls):
        raise NotImplementedError

    @classmethod
    def _get_job_type(cls, _):
        return cls._job_type()

    @classmethod
    def from_job(cls, job):
        if not job.job_type == cls._job_type():
            raise ValueError('wrong job_type: {}'.format(job.job_type))
        if isinstance(job, AbstractSpecificJob):
            job.refresh()
            return job
        else:
            response = cls._client.get(job._job_details_path)
            data = response.json()
            return cls(data)

    @classmethod
    def from_id(cls, project_id, job_id):
        url = cls._job_path(project_id, job_id)
        response = cls._client.get(url, allow_redirects=False)
        data = response.json()
        return cls(data)

    @classmethod
    def get(cls, project_id, job_id):
        # Note: In v3.0 the desired behavior here will be the same as in Job, so we can delete this
        url = cls._job_path(project_id, job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            return cls(data)
        elif response.status_code == 303:
            raise errors.PendingJobFinished
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))


class TrainingPredictionsJob(Job):
    def __init__(self, data, model_id, data_subset, **kwargs):
        super(Job, self).__init__(data, **kwargs)
        self._model_id = model_id
        self._data_subset = data_subset

    @classmethod
    def get(cls, project_id, job_id, model_id=None, data_subset=None):
        """
        Fetches one training predictions job.

        The resulting
        :py:class:`TrainingPredictions <datarobot.models.training_predictions.TrainingPredictions>`
        object will be annotated with `model_id` and `data_subset`.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job id
        model_id : str
            The identifier of the model used for computing training predictions
        data_subset : dr.enums.DATA_SUBSET, optional
            Data subset used for computing training predictions

        Returns
        -------
        job : TrainingPredictionsJob
            The job
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(
            data,
            model_id=model_id,
            data_subset=data_subset,
            completed_resource_url=completed_url,
        )

    def _make_result_from_location(self, location):
        return TrainingPredictions.from_location(
            location,
            model_id=self._model_id,
            data_subset=self._data_subset,
        )

    def refresh(self):
        """
        Update this object with the latest job data from the server.
        """
        data, completed_url = self._data_and_completed_url_for_job(self._this_job_path())
        self.__init__(
            data,
            model_id=self._model_id,
            data_subset=self._data_subset,
            completed_resource_url=completed_url,
        )



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class LiftChart(APIObject):
    """ Lift chart data for model.

    Attributes
    ----------
    source : str
        Lift chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    bins : list of dict
        List of lift chart bins information. Dictionary keys:
            actual : float
                Sum of actual target values in bin
            predicted : float
                Sum of predicted target values in bin
            bin_weight : float
                The weight of the bin. For weighted projects, it is the sum of the weights
                of the rows in the bin. For unweighted projects, it is the number of
                rows in the bin.

    """
    _converter = t.Dict({
        t.Key('source'): t.String,
        t.Key('bins'): t.List(t.Dict({
            t.Key('actual'): t.Float,
            t.Key('predicted'): t.Float,
            t.Key('bin_weight'): t.Float
        }).ignore_extra('*'))
    }).ignore_extra('*')

    def __init__(self, source, bins):
        self.source = source
        self.bins = bins

    def __repr__(self):
        return encode_utf8_if_py2(u'LiftChart({})'.format(self.source))



# coding: utf-8
from datetime import datetime

import dateutil
import pytest
import responses

from datarobot import FrozenModel
from tests.utils import assert_equal_py2, assert_equal_py3


@responses.activate
@pytest.mark.usefixtures('client')
def test_frozen_model_get_returns_valid_data(frozen_json, one_frozen):
    url = 'https://host_name.com/projects/p-id/frozenModels/{}/'.format(one_frozen.id)
    responses.add(responses.GET,
                  url,
                  body=frozen_json,
                  status=200,
                  content_type='application/json')
    frozen = FrozenModel.get('p-id', one_frozen.id)
    assert isinstance(frozen, FrozenModel)
    assert frozen.id == one_frozen.id
    assert frozen.parent_model_id == one_frozen.parent_model_id


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize('sample_pct,training_row_count', [
    (None, None),
    (98.98, None),
    (None, 1000)
])
def test_request_frozen_model(one_model, project_url, frozen_model_job_completed_server_data,
                              sample_pct, training_row_count):
    frozen_models_url = '{}frozenModels/'.format(project_url)
    job_url = '{}modelJobs/{}/'.format(project_url, frozen_model_job_completed_server_data['id'])
    finished_freeze_url = '{}/12344321beefcafe43211234/'.format(frozen_models_url)

    responses.add(responses.POST, frozen_models_url, body='', status=202,
                  content_type='application/json', adding_headers={'Location': job_url})
    responses.add(responses.GET, job_url, json=frozen_model_job_completed_server_data,
                  status=303,
                  adding_headers={'Location': finished_freeze_url})
    if sample_pct:
        model_job = one_model.request_frozen_model(sample_pct=sample_pct)
    elif training_row_count:
        model_job = one_model.request_frozen_model(training_row_count=training_row_count)
    else:
        model_job = one_model.request_frozen_model()
    assert model_job.id == int(frozen_model_job_completed_server_data['id'])


@pytest.fixture
def datetime_frozen_data(datetime_model_data, frozen_data):
    modified_data = dict(frozen_data)
    modified_data.update(datetime_model_data)
    modified_data.pop('sample_pct')
    return modified_data


def test_datetime_frozen_model(datetime_frozen_data):
    mod = FrozenModel.from_data(datetime_frozen_data)
    dt = mod.training_start_date
    assert dt.tzinfo == dateutil.tz.tzutc()
    assert isinstance(mod.training_end_date, datetime)
    assert mod.training_row_count is None
    assert mod.training_duration is None


class TestRepr(object):
    """Tests of method FrozenModel.__repr__"""

    def test_model_type_str(self):
        # when both model_type and id are set
        frozen_model = FrozenModel(id='id', model_type='model_type')
        # then repr contains only model_type
        assert repr(frozen_model) == "FrozenModel('model_type')"

    def test_model_type_unicode(self):
        # when model_type contains non-ascii characters
        frozen_model = FrozenModel(model_type=u'тайп')
        # then repr is an ascii string
        assert_equal_py2(repr(frozen_model), "FrozenModel(u'\\u0442\\u0430\\u0439\\u043f')")
        assert_equal_py3(repr(frozen_model), "FrozenModel('тайп')")

    def test_id(self):
        # when id is set and model_type is missing
        frozen_model = FrozenModel(id='58668cb5bf36cd77a5d8056c')
        # then repr contains id
        assert repr(frozen_model) == "FrozenModel('58668cb5bf36cd77a5d8056c')"



import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class WordCloud(APIObject):
    """ Word cloud data for the model.

    Attributes
    ----------
    ngrams : list of dict
        List of the wordcloud ngrams and corresponding data. Dictionary has following keys:
    ngram : str
        Word or ngram value.
    coefficient : float
        Value from [-1.0, 1.0] range, describes effect of this ngram on the target.
        Large negative value means strong effect toward negative class in classification and
        smaller target value in regression models. Large positive - toward positive class and
        bigger value respectively.
    count : int
        Number of rows in the training sample where this ngram appears.
    frequency : float
        Value from (0.0, 1.0] range, relative frequency of given ngram to most frequent ngram.
    is_stopword : bool
        True for ngrams that DataRobot evaluates as stopwords.
    """
    _converter = t.Dict({
        t.Key('ngrams'): t.List(t.Dict({
            t.Key('ngram'): t.String,
            t.Key('coefficient'): t.Float(gte=-1, lte=1),
            t.Key('count'): t.Int,
            t.Key('frequency'): t.Float(gt=0, lte=1),
            t.Key('is_stopword'): t.Bool
        }).ignore_extra('*'))
    }).ignore_extra('*')

    def __init__(self, ngrams):
        self.ngrams = ngrams

    def __repr__(self):
        return encode_utf8_if_py2(u'WordCloud({} ngrams)'.format(len(self.ngrams)))

    def most_frequent(self, top_n=5):
        """ Return most frequent ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most frequent ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            frequency in descending order.
        """
        return sorted(self.ngrams,
                      key=lambda ngram: ngram['frequency'],
                      reverse=True)[:top_n]

    def most_important(self, top_n=5):
        """ Return most important ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most important ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            absolute coefficient value in descending order.
        """
        return sorted(self.ngrams,
                      key=lambda ngram: abs(ngram['coefficient']),
                      reverse=True)[:top_n]



import pandas
import pytest
import responses

from mock import patch

from datarobot import PredictionExplanationsInitialization, PredictionExplanations
from datarobot.models.prediction_explanations import (PredictionExplanationsPage,
                                                      PredictionExplanationsRow)


@pytest.fixture
def pei_url(project_collection_url, pei_data):
    return '{}{}/models/{}/predictionExplanationsInitialization/'.format(project_collection_url,
                                                                         pei_data['project_id'],
                                                                         pei_data['model_id'])


def test_prediction_explanations_initialization_future_proof(pei_data):
    PredictionExplanationsInitialization.from_data(dict(pei_data, future='new'))


def test_prediction_explanations_future_proof(pe_data):
    PredictionExplanations.from_data(dict(pe_data, future='new'))


def test_prediction_explanations_page_future_proof(pep_data):
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        PredictionExplanationsPage.from_data(dict(pep_data, future='new'))


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_initialization_get_returns_valid_data(pei_url, pei_json, one_pei):
    responses.add(responses.GET,
                  pei_url,
                  body=pei_json,
                  status=200,
                  content_type='application/json')
    initialization = PredictionExplanationsInitialization.get(one_pei.project_id, one_pei.model_id)
    assert isinstance(initialization, PredictionExplanationsInitialization)
    assert initialization.project_id == one_pei.project_id
    assert initialization.model_id == one_pei.model_id
    assert initialization.prediction_explanations_sample


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_initialization_create(pei_data, pei_url, project_url,
                                                       base_job_server_data):
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    responses.add(responses.POST, pei_url,
                  body='', status=202, content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET, job_url,
                  json=base_job_server_data, status=303, adding_headers={'Location': pei_url})

    pei_job = PredictionExplanationsInitialization.create(project_id=pei_data['project_id'],
                                                          model_id=pei_data['model_id'])
    assert str(pei_job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_initialization_delete(pei_url, pei_data):
    responses.add(responses.DELETE, pei_url, status=200)

    pei = PredictionExplanationsInitialization.from_data(pei_data)
    pei.delete()


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_get_returns_valid_data(project_id, project_url, pe_json, one_pe):
    url = '{}predictionExplanationsRecords/{}/'.format(project_url, one_pe.id)
    responses.add(responses.GET,
                  url,
                  body=pe_json,
                  status=200,
                  content_type='application/json')
    record = PredictionExplanations.get(project_id, one_pe.id)
    assert isinstance(record, PredictionExplanations)
    assert record.id == one_pe.id
    assert record.model_id == one_pe.model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_page_get_returns_valid_data(project_id, project_url,
                                                             pep_json, one_pep):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pep.id)
    responses.add(responses.GET,
                  url,
                  body=pep_json,
                  status=200,
                  content_type='application/json')
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        record = PredictionExplanationsPage.get(project_id, one_pep.id)
    assert isinstance(record, PredictionExplanationsPage)
    assert record.id == one_pep.id
    assert record.data == one_pep.data
    assert record.adjustment_method == 'N/A'


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_page_get_returns_valid_data_adjusted(
        project_id, project_url, pep_data_with_adjusted_predictions, one_pep):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pep.id)
    responses.add(responses.GET,
                  url,
                  json=pep_data_with_adjusted_predictions,
                  status=200,
                  content_type='application/json')
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        record = PredictionExplanationsPage.get(project_id, one_pep.id)
    assert isinstance(record, PredictionExplanationsPage)
    assert record.id == one_pep.id
    assert record.data == one_pep.data
    assert record.adjustment_method == 'exposureNormalized'
    for pe_row in record.data:
        assert pe_row['adjusted_prediction'] == pe_row['prediction']
        assert pe_row['adjusted_prediction_values'] == pe_row['prediction_values']


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_create(project_id, project_url, project_collection_url,
                                        base_job_server_data, pep_data):
    pep_base_url = '{}{}/predictionExplanations/'.format(project_collection_url, project_id)
    job_url = '{}jobs/{}/'.format(project_url, base_job_server_data['id'])
    pep_url = '{}/{}/'.format(pep_base_url, pep_data['id'])

    responses.add(responses.POST, pep_base_url,
                  body='', status=202, content_type='application/json',
                  adding_headers={'Location': job_url})
    responses.add(responses.GET, job_url,
                  json=base_job_server_data, status=303, adding_headers={'Location': pep_url})

    pep_job = PredictionExplanations.create(project_id=project_id,
                                            model_id='578e59a41ced2e5a9eb18960',
                                            dataset_id='581c6256100d2b60586980d9')
    assert str(pep_job.id) == base_job_server_data['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_list(project_id, project_url, pe_list_json):
    pe_base_url = '{}predictionExplanationsRecords/'.format(project_url)
    responses.add(responses.GET, pe_base_url + '?modelId=578e59a41ced2e5a9eb18965',
                  status=200, body=pe_list_json, content_type='application/json',
                  match_querystring=True)
    pe_list = PredictionExplanations.list(project_id, model_id='578e59a41ced2e5a9eb18965')
    assert len(pe_list) == 3
    for pe in pe_list:
        assert isinstance(pe, PredictionExplanations)


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_delete(project_collection_url, project_id, pe_data):
    pe_url = '{}{}/predictionExplanationsRecords/{}/'.format(project_collection_url,
                                                             project_id,
                                                             pe_data['id'])
    responses.add(responses.DELETE, pe_url, status=200)

    pe = PredictionExplanations.from_data(pe_data)
    pe.delete()


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_iteration(project_id, project_url, one_pep):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pep.id)
    url_page1 = url + '?offset=0&excludeAdjustedPredictions=true'
    url_page2 = url + '?limit=2&offset=2&excludeAdjustedPredictions=true'
    url_page3 = url + '?limit=2&offset=4&excludeAdjustedPredictions=true'
    resp_data_1 = {'id': one_pep.id,
                   'count': 5,
                   'previous': None,
                   'next': url_page2,
                   'data': one_pep.data[:2],
                   'prediction_explanations_record_location':
                       one_pep.prediction_explanations_record_location}
    resp_data_2 = dict(resp_data_1, previous=url_page1, next=url_page3, data=one_pep.data[2:4])
    resp_data_3 = dict(resp_data_1, previous=url_page2, next=None, data=one_pep.data[4:5])
    responses.add(responses.GET, url_page1, json=resp_data_1, status=200, match_querystring=True)
    responses.add(responses.GET, url_page2, json=resp_data_2, status=200, match_querystring=True)
    responses.add(responses.GET, url_page3, json=resp_data_3, status=200, match_querystring=True)
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        prediction_explanations_rows = list(PredictionExplanations(
            one_pep.id, project_id, '', '', '', 0, 0, ''
        ).get_rows())
    assert len(prediction_explanations_rows) == 5
    for i, row in enumerate(prediction_explanations_rows):
        assert isinstance(row, PredictionExplanationsRow)
        assert row.row_id == one_pep.data[i]['row_id']
        assert row.prediction == one_pep.data[i]['prediction']
        assert row.prediction_values == one_pep.data[i]['prediction_values']
        assert row.prediction_explanations == one_pep.data[i]['prediction_explanations']


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_as_dataframe_regression(project_url, pep_data, one_pe):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data, status=200)
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data['count']
    # 5 columns per prediction explanation, plus prediction and row id
    columns = 5 * one_pe.max_explanations + 2
    assert prediction_explanations_as_dataframe.shape == (rows, columns)


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_as_dataframe_regression_with_exposure(
        project_url, one_pe, pep_data_with_adjusted_predictions):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data_with_adjusted_predictions, status=200)
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data_with_adjusted_predictions['count']
    # 5 columns per prediction explanation, plus prediction, adjusted prediction and row id
    columns = 5 * one_pe.max_explanations + 3
    assert prediction_explanations_as_dataframe.shape == (rows, columns)


@responses.activate
@pytest.mark.usefixtures('client')
def test_prediction_explanations_as_dataframe_classification(project_url,
                                                             pep_data_classification, one_pe):
    url = '{}predictionExplanations/{}/'.format(project_url, one_pe.id)
    responses.add(responses.GET, url, json=pep_data_classification, status=200)
    with patch('datarobot.models.prediction_explanations.t.URL.check',
               side_effect=(lambda url, context=None: url)):
        prediction_explanations_as_dataframe = one_pe.get_all_as_dataframe()
    assert isinstance(prediction_explanations_as_dataframe, pandas.DataFrame)
    rows = pep_data_classification['count']
    # 5 columns per explanation, plus 2 columns per target level, plus row_id and prediction
    columns = (5 * one_pe.max_explanations) + (2 * 2) + 2
    assert prediction_explanations_as_dataframe.shape == (rows, columns)


def test_prediction_explanations_download_to_csv():
    patch_target = ('datarobot.models.prediction_explanations.'
                    'PredictionExplanations.get_all_as_dataframe')
    with patch(patch_target) as pemock:
        prediction_explanations = PredictionExplanations('fake_id', 'fake_project_id', '', '', 3,
                                                         0, 0, '')
        prediction_explanations.download_to_csv('test_file.csv', encoding='ascii')
        pemock.return_value.to_csv.assert_called_once_with(path_or_buf='test_file.csv',
                                                           header=True,
                                                           index=False,
                                                           encoding='ascii')



# -*- encoding: utf-8 -*-

"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import re
from collections import defaultdict
from datetime import datetime, date

import pandas as pd
import six
from dateutil import parser, tz

from .deprecation import deprecated, deprecation_warning  # noqa
from .sourcedata import is_urlsource, recognize_sourcedata, dataframe_to_buffer  # noqa

ALL_CAPITAL = re.compile(r'(.)([A-Z][a-z]+)')
CASE_SWITCH = re.compile(r'([a-z0-9])([A-Z])')
UNDERSCORES = re.compile(r'([a-z]?)(_+)([a-z])')


def underscorize(value):
    partial_result = ALL_CAPITAL.sub(r'\1_\2', value)
    return CASE_SWITCH.sub(r'\1_\2', partial_result).lower()


def underscoreToCamel(match):
    prefix, underscores, postfix = match.groups()
    if len(underscores) > 1:
        # underscoreToCamel('sample_pct__gte') -> 'samplePct__gte'
        return match.group()
    return prefix + postfix.upper()


def camelize(value):
    return UNDERSCORES.sub(underscoreToCamel, value)


def from_api(data, do_recursive=True, keep_attrs=None, keep_null_keys=False):
    if type(data) not in (dict, list):
        return data
    if isinstance(data, list):
        return _from_api_list(data, do_recursive=do_recursive, keep_null_keys=keep_null_keys)
    return _from_api_dict(data, do_recursive=do_recursive,
                          keep_attrs=keep_attrs, keep_null_keys=keep_null_keys)


def _from_api_dict(data, do_recursive=True, keep_attrs=None, keep_null_keys=False):
    keep_attrs = keep_attrs or []
    # prepare attributes in format 'top.middle.bottom' for processing
    parsed_attrs = []
    for attr in keep_attrs:
        if isinstance(attr, six.string_types):
            parsed_attrs.append(attr.split('.'))
        else:
            parsed_attrs.append(attr)
    # take index 0 since recursion goes from top to bottom
    current_level = [attr.pop(0) for attr in parsed_attrs if len(attr)]
    # filter out empty attrs to pass through recursive call
    next_level_attrs = [attr for attr in parsed_attrs if attr]

    app_data = {}
    for k, v in six.iteritems(data):
        k_under = underscorize(k)
        if v is None and k_under not in current_level and not keep_null_keys:
            continue
        if do_recursive:
            data_val = from_api(v, do_recursive=do_recursive,
                                keep_attrs=next_level_attrs, keep_null_keys=keep_null_keys)
        else:
            data_val = v
        app_data[k_under] = data_val
    return app_data


def _from_api_list(data, do_recursive=True, keep_null_keys=False):
    return [from_api(datum, do_recursive=do_recursive, keep_null_keys=keep_null_keys)
            for datum in data]


def remove_empty_keys(metadata, keep_attrs=None):
    keep_attrs = keep_attrs or []
    return {k: v for k, v in metadata.items() if v is not None or k in keep_attrs}


def parse_time(time_str):
    try:
        return parser.parse(time_str, tzinfos={'UTC': tz.tzutc()})
    except Exception:
        return time_str


def datetime_to_string(datetime_obj):
    """ Converts to isoformat after dropping microseconds
    """
    if not isinstance(datetime_obj, datetime):
        msg = 'expected to be passed a datetime.datetime, was passed {}'.format(type(datetime_obj))
        raise ValueError(msg)
    stripped = datetime_obj.replace(microsecond=0)
    return stripped.isoformat()


def to_api(data, keep_attrs=None):
    """
    :param data: dictionary {'max_digits': 1}
    :return: {'maxDigits': 1}
    """
    if not data:
        return {}
    assert isinstance(data, dict), 'Wrong type'
    return _to_api_item(data, keep_attrs)


def _to_api_item(item, keep_attrs=None):
    if isinstance(item, dict):
        dense_item = remove_empty_keys(item, keep_attrs)
        return {camelize(k): _to_api_item(v) for k, v in six.iteritems(dense_item)}
    elif isinstance(item, list):
        return [_to_api_item(subitem) for subitem in item]
    elif isinstance(item, datetime):
        return datetime_to_string(item)
    elif isinstance(item, date):
        return item.isoformat()
    else:
        return item


def get_id_from_response(response):
    location_string = response.headers['Location']
    return get_id_from_location(location_string)


def get_id_from_location(location_string):
    return location_string.split('/')[-2]


def get_duplicate_features(features):
    duplicate_features = set()
    seen_features = set()
    for feature in features:
        if feature in seen_features:
            duplicate_features.add(feature)
        else:
            seen_features.add(feature)
    return list(duplicate_features)


def raw_prediction_response_to_dataframe(pred_response, class_prefix):
    """Raw predictions for classification come as nested json objects.

    This will extract it to be a single dataframe.

    Parameters
    ----------
    pred_response : dict
        The loaded json object returned from the prediction route.

    Returns
    -------
    frame : pandas.DataFrame

    """
    predictions = from_api(pred_response)['predictions']
    frame = pd.DataFrame.from_records(predictions)
    if 'prediction_values' in frame.columns:
        frame = _pivot_prediction_labels(frame, class_prefix)
    return frame


def _pivot_prediction_labels(frame, class_prefix):
    """ Pivot the prediciton_values from the dictionary to columns

    Parameters
    ----------
    frame : pandas.DataFrame

    Returns
    -------
    frame : pandas.DataFrame

    """
    wrapper = defaultdict(list)
    for pred_row in frame['prediction_values']:
        for pred_value in pred_row:
            col_name = u''.join((class_prefix, u'{}'.format(pred_value['label'])))
            wrapper[col_name].append(pred_value['value'])
    pred_frame = pd.DataFrame.from_records(wrapper)
    frame = pd.concat([frame, pred_frame], axis=1)
    return frame.drop('prediction_values', 1)


def encode_utf8_if_py2(string):
    """__repr__ is supposed to return string (bytes) in 2 and string (unicode) in 3
    this function can be used to convert our unicode to strings in Python 2 but leave them alone
    in Python 3"""
    return string.encode('utf-8') if six.PY2 else string



import pytest

from datarobot import Client
from datarobot.client import get_client


def test_token_endpoint_overrides_config_path(config_path):
    Client(token='code_token', endpoint='https://code_endpoint.com', config_path=config_path)
    client = get_client()
    assert client.token == 'code_token'
    assert client.endpoint == 'https://code_endpoint.com'


@pytest.mark.usefixtures('env_config_path')
def test_config_overrides_env_variables(config_path, code_config_token,
                                        code_config_endpoint):
    Client(config_path=config_path)
    client = get_client()
    assert client.token == code_config_token
    assert client.endpoint == code_config_endpoint


@pytest.mark.usefixtures('env_config_path', 'env_config_vars', 'no_version_check')
def test_env_config_overrides_env_vars(env_config_endpoint,
                                       env_config_token):
    client = get_client()
    assert client.token == env_config_token
    assert client.endpoint == env_config_endpoint


@pytest.mark.usefixtures('env_config_vars', 'default_config', 'no_version_check')
def test_config_vars_override_config_file_in_default_location(env_token, env_endpoint):
    client = get_client()
    assert client.token == env_token
    assert client.endpoint == env_endpoint


@pytest.mark.usefixtures('default_config', 'no_version_check')
def test_default_config_if_nothing_else(default_config_endpoint, default_config_token):
    client = get_client()
    assert client.token == default_config_token
    assert client.endpoint == default_config_endpoint


def test_client_ssl_verify_default_true():
    client = Client(token='token', endpoint='https://code_endpoint.com')
    assert client.verify is True


def test_client_ssl_verify_false():
    client = Client(token='token', endpoint='https://code_endpoint.com', ssl_verify=False)
    assert client.verify is False



import json

import pytest
import responses

import datarobot as dr
from datarobot.enums import POSTGRESQL_DRIVER


class TestCreateMysql(object):
    @pytest.fixture
    def mysql_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = '{}/{}/'.format(unittest_endpoint, 'mysqlProjects')
        responses.add(
            responses.POST,
            mysql_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @pytest.fixture
    def string_encryption_response(self, unittest_endpoint):
        string_encrypt_url = '{}/stringEncryptions/'.format(unittest_endpoint)
        responses.add(
            responses.POST,
            string_encrypt_url,
            body=json.dumps({'cipherText': 'cipherpass'}),
            status=200,
            content_type='application/json',
        )

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'mysql_create_response',
                             'known_warning')
    def test_create_basic(self):
        dr.Project.create_from_mysql(server='127.0.0.1',
                                     database='test_database',
                                     table='test_table',
                                     user='test_user')
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['server'] == '127.0.0.1'
        assert create_body['database'] == 'test_database'
        assert create_body['table'] == 'test_table'
        assert create_body['user'] == 'test_user'
        assert 'port' not in create_body  # rely on default on server if not specified

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'mysql_create_response',
                             'known_warning')
    def test_create_with_all(self):
        dr.Project.create_from_mysql(server='127.0.0.1',
                                     database='test_database',
                                     table='test_table',
                                     user='test_user',
                                     port=3307,
                                     prefetch=10000,
                                     project_name='A MySQL Project',
                                     encrypted_password='ciphertext',
                                     max_wait=100)
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['server'] == '127.0.0.1'
        assert create_body['database'] == 'test_database'
        assert create_body['table'] == 'test_table'
        assert create_body['user'] == 'test_user'
        assert create_body['port'] == 3307
        assert create_body['prefetch'] == 10000
        assert create_body['projectName'] == 'A MySQL Project'
        assert create_body['encryptedPassword'] == 'ciphertext'

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'string_encryption_response',
                             'project_creation_responses',
                             'mysql_create_response',
                             'known_warning')
    def test_create_with_plain_password(self):
        dr.Project.create_from_mysql(server='127.0.0.1',
                                     database='test_database',
                                     table='test_table',
                                     user='test_user',
                                     password='plaintext')
        create_body = json.loads(responses.calls[1].request.body)
        assert create_body['server'] == '127.0.0.1'
        assert create_body['database'] == 'test_database'
        assert create_body['table'] == 'test_table'
        assert create_body['user'] == 'test_user'
        assert create_body['encryptedPassword'] != 'plaintext'

    @pytest.mark.usefixtures('known_warning')
    def test_password_and_encrypted_password_is_error(self):
        with pytest.raises(ValueError):
            dr.Project.create_from_mysql(server='127.0.0.1',
                                         database='test_database',
                                         table='test_table',
                                         user='test_user',
                                         password='plaintext',
                                         encrypted_password='ciphertext')


class TestCreatePostgreSQL(object):
    @pytest.fixture
    def postgresql_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = '{}/{}/'.format(unittest_endpoint, 'postgresqlProjects')
        responses.add(
            responses.POST,
            mysql_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'postgresql_create_response',
                             'known_warning')
    def test_create_basic(self):
        dr.Project.create_from_postgresql(server='127.0.0.1',
                                          database='test_database',
                                          table='test_table',
                                          username='test_user')
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['server'] == '127.0.0.1'
        assert create_body['database'] == 'test_database'
        assert create_body['table'] == 'test_table'
        assert create_body['username'] == 'test_user'
        # rely for these on default on server if not specified
        assert 'port' not in create_body
        assert 'driver' not in create_body
        assert 'fetch' not in create_body
        assert 'useDeclareFetch' not in create_body

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'postgresql_create_response',
                             'known_warning')
    def test_create_with_all(self):
        dr.Project.create_from_postgresql(server='127.0.0.1',
                                          database='test_database',
                                          table='test_table',
                                          username='test_user',
                                          port=3307,
                                          fetch=10000,
                                          use_declare_fetch=False,
                                          project_name='A PostgreSQL Project',
                                          encrypted_password='ciphertext',
                                          driver=POSTGRESQL_DRIVER.UNICODE,
                                          max_wait=100)
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['server'] == '127.0.0.1'
        assert create_body['database'] == 'test_database'
        assert create_body['table'] == 'test_table'
        assert create_body['username'] == 'test_user'
        assert create_body['port'] == 3307
        assert create_body['fetch'] == 10000
        assert not create_body['useDeclareFetch']
        assert create_body['projectName'] == 'A PostgreSQL Project'
        assert create_body['encryptedPassword'] == 'ciphertext'
        assert create_body['driver'] == POSTGRESQL_DRIVER.UNICODE

    @pytest.mark.usefixtures('known_warning')
    def test_password_and_encrypted_password_is_error(self):
        with pytest.raises(ValueError):
            dr.Project.create_from_postgresql(server='127.0.0.1',
                                              database='test_database',
                                              table='test_table',
                                              username='test_user',
                                              password='plaintext',
                                              encrypted_password='ciphertext')


class TestCreateOracle(object):
    @pytest.fixture
    def oracle_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = '{}/{}/'.format(unittest_endpoint, 'oracleProjects')
        responses.add(
            responses.POST,
            mysql_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'oracle_create_response',
                             'known_warning')
    def test_create_basic(self):
        dr.Project.create_from_oracle(dbq='127.0.0.1:1111/xe',
                                      table='test_table',
                                      username='test_user')
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['dbq'] == '127.0.0.1:1111/xe'
        assert create_body['table'] == 'test_table'
        assert create_body['username'] == 'test_user'
        # rely for this on default on server if not specified
        assert 'fetch_buffer_size' not in create_body

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'oracle_create_response',
                             'known_warning')
    def test_create_with_all(self):
        dr.Project.create_from_oracle(dbq='127.0.0.1:1111/xe',
                                      table='test_table',
                                      username='test_user',
                                      project_name='A Oracle Project',
                                      encrypted_password='ciphertext',
                                      fetch_buffer_size=10000,
                                      max_wait=100)
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['dbq'] == '127.0.0.1:1111/xe'
        assert create_body['table'] == 'test_table'
        assert create_body['username'] == 'test_user'
        assert create_body['fetchBufferSize'] == 10000
        assert create_body['projectName'] == 'A Oracle Project'
        assert create_body['encryptedPassword'] == 'ciphertext'

    @pytest.mark.usefixtures('known_warning')
    def test_password_and_encrypted_password_is_error(self):
        with pytest.raises(ValueError):
            dr.Project.create_from_oracle(dbq='127.0.0.1:1111/xe',
                                          table='test_table',
                                          username='test_user',
                                          password='plaintext',
                                          encrypted_password='ciphertext')


class TestHDFSCreate(object):
    @pytest.fixture
    def hdfs_create_response(self, unittest_endpoint, async_url):
        mysql_create_url = '{}/{}/'.format(unittest_endpoint, 'hdfsProjects')
        responses.add(
            responses.POST,
            mysql_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'hdfs_create_response')
    def test_create_hdfs_minimum(self):
        dr.Project.create_from_hdfs(url='hdfs://webhdfs.com/directory/file.csv')
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['url'] == 'hdfs://webhdfs.com/directory/file.csv'

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'hdfs_create_response')
    def test_create_hdfs_all_options(self):
        dr.Project.create_from_hdfs(url='hdfs://webhdfs.com/directory/file.csv',
                                    port=9999,
                                    project_name='HDFS Project',
                                    max_wait=100)
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body['url'] == 'hdfs://webhdfs.com/directory/file.csv'
        assert create_body['port'] == 9999
        assert create_body['projectName'] == 'HDFS Project'


class TestDataSourceCreate(object):
    @pytest.fixture
    def data_source_create_response(self, unittest_endpoint, async_url):
        project_create_url = '{}/{}'.format(unittest_endpoint, dr.Project._path)
        responses.add(
            responses.POST,
            project_create_url,
            body='',
            status=202,
            content_type='application/json',
            adding_headers={'Location': async_url}
        )

    @pytest.fixture
    def data_source(self):
        return dr.DataSource('5acc8437ec8d670001ba16bf')

    @responses.activate
    @pytest.mark.usefixtures('client',
                             'project_creation_responses',
                             'data_source_create_response')
    def test_create_from_data_source_id(self):
        username = 'test_user'
        password = 'megasuperubersecretpassword'
        data_source_id = '5acc8437ec8d670001ba16bf'
        project_name = 'data source project'
        dr.Project.create_from_data_source(data_source_id, username, password,
                                           project_name=project_name)
        create_body = json.loads(responses.calls[0].request.body)
        assert create_body == {
            'user': username,
            'password': password,
            'dataSourceId': data_source_id,
            'projectName': project_name
        }



import json

from mock import patch
import pytest
import responses

from datarobot import Project, AppPlatformError
from datarobot import errors
from datarobot.enums import DEFAULT_TIMEOUT
from tests.utils import SDKTestcase, assert_raised_regex


class Test400LevelErrors(SDKTestcase):

    @responses.activate
    def test_401_error_message(self):
        """For whatever reason, the `requests` library completely ignores the
        body in a 401 error, so, here is our own message
        """
        responses.add(
            responses.GET,
            'https://host_name.com/projects/',
            status=401,
        )

        with pytest.raises(AppPlatformError) as exc_info:
            Project.list()
        assert_raised_regex(exc_info, 'not properly authenticated')

    @responses.activate
    def test_403_error_message(self):
        """ Check that we output same error message that server returns on 403
        """
        error_text = 'Some permission error'
        error_response = {'message': error_text}

        responses.add(
            responses.GET,
            'https://host_name.com/projects/',
            status=403,
            body=json.dumps(error_response)
        )

        with pytest.raises(AppPlatformError) as exc_info:
            Project.list()
        expected_error = '{} client error: {}'.format(403, json.dumps(error_response))
        assert str(exc_info.value) == expected_error

    @responses.activate
    def test_model_already_added_exception(self):
        responses.add(
            responses.POST,
            'https://host_name.com/projects/p-id/models/',
            status=422,
            body=json.dumps({'message': u'Model already added', 'errorName': u'JobAlreadyAdded'})
        )

        with pytest.raises(errors.JobAlreadyRequested):
            Project('p-id').train('some-blueprint-id')

    @responses.activate
    def test_extracts_error_message(self):
        data = {'message': 'project p-id has been deleted'}
        responses.add(responses.GET, 'https://host_name.com/projects/p-id/models/',
                      status=422,
                      body=json.dumps(data),
                      content_type='application/json')

        with pytest.raises(AppPlatformError) as exc_info:
            Project('p-id').get_models()
        assert_raised_regex(exc_info, 'project p-id has been deleted')


def test_client_uses_both_timeouts(client):
    connect_timeout = 33
    client.connect_timeout = connect_timeout

    with patch('datarobot.rest.requests.Session.request') as request_mock:
        client.get('ping')
        args, kwargs = request_mock.call_args
        called_timeout = kwargs['timeout']
        assert called_timeout == (connect_timeout, DEFAULT_TIMEOUT.READ)


@responses.activate
def test_client_passes_read_timeout_from_build_request(client, temporary_file):
    connect_timeout = 33
    read_timeout = 180
    client.connect_timeout = connect_timeout

    with patch('datarobot.rest.requests.Session.request') as request_mock:
        client.build_request_with_file('POST', 'file_upload', 'my_file.csv',
                                       file_path=temporary_file, read_timeout=read_timeout)
        args, kwargs = request_mock.call_args
        called_timeout = kwargs['timeout']
        assert called_timeout == (connect_timeout, read_timeout)



import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.models.api_object import APIObject

from ..utils import encode_utf8_if_py2, from_api, parse_time


_data_source_params_converter = t.Dict({
    t.Key('data_store_id'): t.String() | t.Null,
    t.Key('table', optional=True): t.String() | t.Null,
    t.Key('schema', optional=True): t.String() | t.Null,
    t.Key('partition_column', optional=True): t.String() | t.Null,
    t.Key('query', optional=True): t.String() | t.Null,
    t.Key('fetch_size', optional=True): t.Int() | t.Null
}).ignore_extra('*')


class DataSourceParameters(object):
    """ Data request configuration

    Attributes
    ----------
    data_store_id : str
        the id of the DataStore.
    table : str
        optional, the name of specified database table.
    schema : str
        optional, the name of the schema associated with the table.
    partition_column : str
        optional, the name of the partition column.
    query : str
        optional, the user specified SQL query.
    fetch_size : int
        optional, a user specified fetch size in the range [1, 20000].
        By default a fetchSize will be assigned to balance throughput and memory usage
    """
    def __init__(self, data_store_id=None, table=None, schema=None, partition_column=None,
                 query=None, fetch_size=None):
        _data_source_params_converter.check({
            'data_store_id': data_store_id,
            'table': table,
            'schema': schema,
            'partition_column': partition_column,
            'query': query,
            'fetch_size': fetch_size
        })
        self.data_store_id = data_store_id
        self.table = table
        self.schema = schema
        self.partition_column = partition_column
        self.query = query
        self.fetch_size = fetch_size

    def collect_payload(self):
        return {
            'data_store_id': self.data_store_id,
            'table': self.table,
            'schema': self.schema,
            'partition_column': self.partition_column,
            'query': self.query,
            'fetch_size': self.fetch_size
        }

    @classmethod
    def from_server_data(cls, data):
        converted_data = _data_source_params_converter.check(from_api(data))
        return cls(**converted_data)

    def __eq__(self, other):
        self_payload = self.collect_payload()
        other_payload = other.collect_payload()
        del self_payload['data_store_id']
        del other_payload['data_store_id']
        return self_payload == other_payload


class DataSource(APIObject):
    """ A data source. Represents data request

    Attributes
    ----------
    data_source_id : str
        the id of the data source.
    data_source_type : str
        the type of data source.
    canonical_name : str
        the user-friendly name of the data source.
    creator : str
        the id of the user who created the data source.
    updated : datetime.datetime
        the time of the last update.
    params : DataSourceParameters
        a list specifying data source parameters.
    """
    _path = 'externalDataSources/'
    _client = staticproperty(get_client)
    _converter = t.Dict({
        t.Key('id', optional=True) >> 'data_source_id': t.String(),
        t.Key('type') >> 'data_source_type': t.String(),
        t.Key('canonical_name'): t.String(),
        t.Key('creator'): t.String(),
        t.Key('params'): _data_source_params_converter,
        t.Key('updated'): parse_time,
    }).ignore_extra('*')

    def __init__(self, data_source_id=None, data_source_type=None, canonical_name=None,
                 creator=None, updated=None, params=None):
        self._id = data_source_id
        self._type = data_source_type
        self.canonical_name = canonical_name
        self._creator = creator
        self._updated = updated
        self.params = params

    @classmethod
    def list(cls):
        """
        Returns list of available data sources.

        Returns
        -------
        data_sources : list of DataSource instances
            contains a list of available data sources.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_sources = dr.DataSource.list()
            >>> data_sources
            [DataSource('Diagnostics'), DataSource('Airlines 100mb'), DataSource('Airlines 10mb')]
        """
        r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data['data']]

    @classmethod
    def get(cls, data_source_id):
        """
        Gets the data source.

        Parameters
        ----------
        data_source_id : str
            the identifier of the data source.

        Returns
        -------
        data_source : DataSource
            the requested data source.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_source = dr.DataSource.get('5a8ac9ab07a57a0001be501f')
            >>> data_source
            DataSource('Diagnostics')
        """
        return cls.from_location('{}{}/'.format(cls._path, data_source_id))

    @classmethod
    def create(cls, data_source_type, canonical_name, params):
        """
        Creates the data source.

        Parameters
        ----------
        data_source_type : str
            the type of data source.
        canonical_name : str
            the user-friendly name of the data source.
        params : DataSourceParameters
            a list specifying data source parameters.

        Returns
        -------
        data_source : DataSource
            the created data source.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> params = dr.DataSourceParameters(
            ...     data_store_id='5a8ac90b07a57a0001be501e',
            ...     query='SELECT * FROM airlines10mb WHERE "Year" >= 1995;'
            ... )
            >>> data_source = dr.DataSource.create(
            ...     data_source_type='jdbc',
            ...     canonical_name='airlines stats after 1995',
            ...     params=params
            ... )
            >>> data_source
            DataSource('airlines stats after 1995')
        """
        payload = {
            'type': data_source_type,
            'canonicalName': canonical_name,
            'params': params.collect_payload()
        }
        return cls.from_server_data(
            cls._client.post(cls._path, data=payload).json()
        )

    def update(self, canonical_name=None, params=None):
        """
        Creates the data source.

        Parameters
        ----------
        canonical_name : str
            optional, the user-friendly name of the data source.
        params : DataSourceParameters
            optional, the identifier of the DataDriver.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_source = dr.DataSource.get('5ad840cc613b480001570953')
            >>> data_source
            DataSource('airlines stats after 1995')
            >>> params = dr.DataSourceParameters(
            ...     query='SELECT * FROM airlines10mb WHERE "Year" >= 1990;'
            ... )
            >>> data_source.update(
            ...     canonical_name='airlines stats after 1990',
            ...     params=params
            ... )
            >>> data_source
            DataSource('airlines stats after 1990')
        """
        payload = {
            'canonicalName': canonical_name or self.canonical_name,
            'params': params.collect_payload() if params else self.params.collect_payload()
        }
        r_data = self._client.patch('{}{}/'.format(self._path, self.id), data=payload).json()
        self.canonical_name = r_data['canonicalName']
        self.params = DataSourceParameters.from_server_data(r_data.pop('params'))

    def delete(self):
        """ Removes the DataSource """
        self._client.delete('{}{}/'.format(self._path, self.id))

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        converted_data = cls._converter.check(from_api(data))
        params = converted_data.pop('params')
        data_store_id = params.pop('data_store_id')
        converted_data['params'] = DataSourceParameters(data_store_id, **params)
        return cls(**converted_data)

    def __repr__(self):
        return encode_utf8_if_py2(u'{}(\'{}\')'.format(self.__class__.__name__,
                                                       self.canonical_name or self.id))

    @property
    def id(self):
        return self._id

    @property
    def creator(self):
        return self._creator

    @property
    def type(self):
        return self._type

    @property
    def updated(self):
        return self._updated



def unpaginate(initial_url, initial_params, client):
    """ Iterate over a paginated endpoint and get all results

    Assumes the endpoint follows the "standard" pagination interface (data stored under "data",
    "next" used to link next page, "offset" and "limit" accepted as query parameters).

    Yields
    ------
    data : dict
        a series of objects from the endpoint's data, as raw server data
    """
    resp_data = client.get(initial_url, params=initial_params).json()
    for item in resp_data['data']:
        yield item
    while resp_data['next'] is not None:
        next_url = resp_data['next']
        resp_data = client.get(next_url).json()
        for item in resp_data['data']:
            yield item



import json
import pytest
import responses

from datarobot import ImportedModel
from tests.test_helpers import fixture_file_path

imported_model_dict = {u'originUrl': u'http://app.datarobot.com/',
                       u'datasetName': u'train_data.csv', u'modelName': u'Linear Regression (5)',
                       u'displayName': u'Linear Regression (5)', u'target': u'target',
                       u'projectId': u'57b37876c808912b1c87094a',
                       u'importedByUsername': u'david.chudzicki@datarobot.com',
                       u'note': None, u'importedAt': u'2016-08-28 22:39:56.117000', u'version': 1.8,
                       u'projectName': u'Unnamed Project', u'featurelistName': u'week_day',
                       u'createdByUsername': u'david.chudzicki@datarobot.com',
                       u'modelId': u'57b70cf5c80891295500c6e8',
                       u'importedById': u'56a5a34b16f4b622fb8426f6',
                       u'id': u'301d0dd18e034c058dbed3bdf8207a4a',
                       u'createdById': u'56a5a34b16f4b622fb8426f6'}


def add_imported_model_get_response(model_id):
    url = 'https://host_name.com/importedModels/{}/'.format(model_id)
    responses.add(responses.GET, url, body=json.dumps(imported_model_dict))


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.fixture
def imported_model():
    add_imported_model_get_response(imported_model_dict['id'])
    return ImportedModel.get(imported_model_dict['id'])


@responses.activate
@pytest.mark.usefixtures('client')
def test_get(imported_model):
    assert imported_model.id == imported_model_dict['id']


@responses.activate
@pytest.mark.usefixtures('client')
def test_list():
    response_dict = {u'count': 1, u'next': None, u'previous': None, u'data': [imported_model_dict]}
    url = 'https://host_name.com/importedModels/'
    responses.add(responses.GET, url, body=json.dumps(response_dict))
    imported_models = ImportedModel.list(imported_model_dict['id'])
    assert isinstance(imported_models, list)
    assert len(imported_models) == 1
    assert isinstance(imported_models[0], ImportedModel)


@responses.activate
@pytest.mark.usefixtures('client')
def test_update(imported_model):
    url = 'https://host_name.com/importedModels/{}/'.format(imported_model_dict['id'])
    responses.add(responses.PATCH, url, body='this is not actually used')
    new_note = 'new note'
    new_display_name = 'new_display_name'
    imported_model.update(display_name=new_display_name, note=new_note)
    assert imported_model.note == new_note
    assert imported_model.display_name == new_display_name


@responses.activate
@pytest.mark.usefixtures('client')
def test_delete(imported_model):
    url = 'https://host_name.com/importedModels/{}/'.format(imported_model_dict['id'])
    responses.add(responses.DELETE, url, body='not used')
    imported_model.delete()


@responses.activate
@pytest.mark.usefixtures('client')
def test_create():
    responses.add(responses.POST,
                  'https://host_name.com/importedModels/',
                  body='',
                  status=202,
                  content_type='application/json',
                  adding_headers={'Location': 'https://host_name.com/status/status-id/'}
                  )

    responses.add(responses.GET,
                  'https://host_name.com/status/status-id/',
                  status=303,
                  body='',
                  content_type='application/json',
                  adding_headers={
                      'Location': 'https://host_name.com/importedModels/{}/'.format(
                          imported_model_dict['id'])}
                  )
    add_imported_model_get_response(imported_model_dict['id'])

    imported_model = ImportedModel.create(fixture_file_path('some_model.drmodel'))
    assert isinstance(imported_model, ImportedModel)



import json

import pytest
import responses

from datarobot.models.model import Model
from datarobot.models.ruleset import Ruleset


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_prime_models(project_url, project,
                           prime_model_server_data, ruleset_with_model_server_data):
    response_data = {'count': 1, 'previous': None, 'next': None,
                     'data': [prime_model_server_data]}
    responses.add(responses.GET, '{}primeModels/'.format(project_url),
                  body=json.dumps(response_data))

    prime_models = project.get_prime_models()

    assert len(prime_models) == 1
    expected_model = Model.from_server_data(prime_model_server_data)
    prime_model = prime_models[0]
    assert prime_model.id == expected_model.id
    assert prime_model.model_category == expected_model.model_category
    assert prime_model.model_type == expected_model.model_type

    expected_ruleset = Ruleset.from_server_data(ruleset_with_model_server_data)
    assert prime_model.ruleset.id == expected_ruleset.id
    assert prime_model.parent_model_id == expected_ruleset.parent_model_id



import trafaret as t

from .. import errors
from ..enums import JOB_TYPE, PREDICT_JOB_STATUS, DEFAULT_MAX_WAIT, PREDICTION_PREFIX
from .job import AbstractSpecificJob
from ..utils import (
    deprecated,
    raw_prediction_response_to_dataframe,
    recognize_sourcedata,
    get_id_from_response,
    from_api,
    retry,
)


def wait_for_async_predictions(project_id, predict_job_id, max_wait=DEFAULT_MAX_WAIT):
    """
    Given a Project id and PredictJob id poll for status of process
    responsible for predictions generation until it's finished

    Parameters
    ----------
    project_id : str
        The identifier of the project
    predict_job_id : str
        The identifier of the PredictJob
    max_wait : int, optional
        Time in seconds after which predictions creation is considered
        unsuccessful

    Returns
    -------
    predictions : pandas.DataFrame
        Generated predictions.

    Raises
    ------
    AsyncPredictionsGenerationError
        Raised if status of fetched PredictJob object is ``error``
    AsyncTimeoutError
        Predictions weren't generated in time, specified by ``max_wait``
        parameter
    """
    # Note: Don't need this in 3.0. (Use `Job.get_result_when_complete` method instead.)
    for index in retry.wait(max_wait):
        try:
            predict_job = PredictJob.get(project_id, predict_job_id)
        except errors.PendingJobFinished:
            return PredictJob.get_predictions(project_id, predict_job_id)
        if predict_job.status in (PREDICT_JOB_STATUS.ERROR, PREDICT_JOB_STATUS.ABORTED):
            e_msg = 'Predictions generation unsuccessful'
            raise errors.AsyncPredictionsGenerationError(e_msg)

    timeout_msg = 'Predictions generation timed out in {} seconds'.format(
        max_wait)
    raise errors.AsyncTimeoutError(timeout_msg)


class PredictJob(AbstractSpecificJob):

    """ Tracks asynchronous work being done within a project

    Attributes
    ----------
    id : int
        the id of the job
    project_id : str
        the id of the project the job belongs to
    status : str
        the status of the job - will be one of ``datarobot.enums.QUEUE_STATUS``
    job_type : str
        what kind of work the job is doing - will be 'predict' for predict jobs
    is_blocked : bool
        if true, the job is blocked (cannot be executed) until its dependencies are resolved
    message : str
        a message about the state of the job, typically explaining why an error occurred
    """

    _extra_fields = frozenset(['message'])
    _converter_extra = t.Dict({
        t.Key('message', optional=True, default=''): t.String(allow_blank=True),
    })

    def __init__(self, data, completed_resource_url=None):
        super(PredictJob, self).__init__(data, completed_resource_url=completed_resource_url)
        from . import Model
        data = from_api(data)
        self.model = Model(project_id=data.get('project_id'), id=data.get('model_id'))

    def __repr__(self):
        return 'PredictJob({}, status={})'.format(self.model,
                                                  self.status)  # pragma:no cover

    @classmethod
    def _job_type(cls):
        return JOB_TYPE.PREDICT

    @classmethod
    def from_job(cls, job):
        """ Transforms a generic Job into a PredictJob

        Parameters
        ----------
        job: Job
            A generic job representing a PredictJob

        Returns
        -------
        predict_job: PredictJob
            A fully populated PredictJob with all the details of the job

        Raises
        ------
        ValueError:
            If the generic Job was not a predict job, e.g. job_type != JOB_TYPE.PREDICT
        """
        return super(PredictJob, cls).from_job(job)

    @classmethod
    @deprecated(deprecated_since_version='v2.3', will_remove_version='v3.0',
                message='Please use Model.request_predictions instead.')
    def create(cls, model, sourcedata):
        """
        .. note::
           Deprecated in v2.3 in favor of ``Project.upload_dataset`` and
           ``Model.request_predictions``. That workflow allows you to reuse the same dataset
           for predictions from multiple models within one project.

        Starts predictions generation for provided data
        using previously created model.

        Parameters
        ----------
        model : Model
            Model to use for predictions generation
        sourcedata : str, file or pandas.DataFrame
            Data to be used for predictions.
            If this parameter is a str, it can be either a path to a local file
            or raw file content. If using a file on disk, the filename must
            consist of ASCII characters only. The file must be a CSV, and cannot
            be compressed

        Returns
        -------
        predict_job_id : str
            id of created job, can be used as parameter to ``PredictJob.get``
            or ``PredictJob.get_predictions`` methods or
            ``wait_for_async_predictions`` function

        Raises
        ------
        InputNotUnderstoodError
            If the parameter for sourcedata didn't resolve into known data types

        Examples
        --------
        .. code-block:: python

            model = Model.get('p-id', 'l-id')
            predict_job = PredictJob.create(model, './data_to_predict.csv')
        """
        url = 'projects/{}/predictions/'.format(model.project.id)
        form_data = {'model_id': model.id}
        default_fname = 'predict.csv'
        filesource_kwargs = recognize_sourcedata(sourcedata, default_fname)

        response = cls._client.build_request_with_file(
            url=url,
            form_data=form_data,
            method='post',
            **filesource_kwargs)
        return get_id_from_response(response)

    @classmethod
    def _job_path(cls, project_id, job_id):
        return 'projects/{}/predictJobs/{}/'.format(project_id, job_id)

    @classmethod
    def get(cls, project_id, predict_job_id):
        """
        Fetches one PredictJob. If the job finished, raises PendingJobFinished
        exception.

        Parameters
        ----------
        project_id : str
            The identifier of the project the model on which prediction
            was started belongs to
        predict_job_id : str
            The identifier of the predict_job

        Returns
        -------
        predict_job : PredictJob
            The pending PredictJob

        Raises
        ------
        PendingJobFinished
            If the job being queried already finished, and the server is
            re-routing to the finished predictions.
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        return super(PredictJob, cls).get(project_id, predict_job_id)

    @classmethod
    def get_predictions(cls, project_id, predict_job_id, class_prefix=PREDICTION_PREFIX.DEFAULT):
        """
        Fetches finished predictions from the job used to generate them.

        .. note::
            The prediction API for classifications now returns an additional prediction_values
            dictionary that is converted into a series of class_prefixed columns in the final
            dataframe. For example, <label> = 1.0 is converted to 'class_1.0'. If you are on an
            older version of the client (prior to v2.8), you must update to v2.8 to correctly pivot
            this data.

        Parameters
        ----------
        project_id : str
            The identifier of the project to which belongs the model used
            for predictions generation
        predict_job_id : str
            The identifier of the predict_job
        class_prefix : str
            The prefix to append to labels in the final dataframe (e.g., apple -> class_apple)

        Returns
        -------
        predictions : pandas.DataFrame
            Generated predictions

        Raises
        ------
        JobNotFinished
            If the job has not finished yet
        AsyncFailureError
            Querying the predict_job in question gave a status code other than 200 or 303
        """
        # Note: Don't need this in 3.0. (Use `Job.get_result` method instead.)
        url = cls._job_path(project_id, predict_job_id)
        response = cls._client.get(url, allow_redirects=False)

        if response.status_code == 200:
            data = response.json()
            raise errors.JobNotFinished('Pending job status is {}'.format(data['status']))
        elif response.status_code == 303:
            location = response.headers['Location']
            response = cls._client.get(location, join_endpoint=False)
            return raw_prediction_response_to_dataframe(response.json(), class_prefix)
        else:
            e_msg = 'Server unexpectedly returned status code {}'
            raise errors.AsyncFailureError(
                e_msg.format(response.status_code))



import pytest

from datarobot import Job


@pytest.mark.usefixtures('client')
@pytest.mark.parametrize('is_blocked', [False, True])
def test_job_initialization_with_is_blocked(is_blocked):
    job = Job({
        'status': 'queue',
        'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
        'id': '1',
        'jobType': 'model',
        'projectId': 'p-id',
        'isBlocked': is_blocked,
    })

    assert job.id == 1
    assert job.status == 'queue'
    assert job.job_type == 'model'
    assert job.project_id == 'p-id'
    assert job.is_blocked is is_blocked



import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.models.api_object import APIObject


class Solution(object):
    """Eureqa Solution.

    A solution represents a possible Eureqa model; however not all solutions
    have models associated with them.  It must have a model created before
    it can be used to make predictions, etc.

    Attributes
    ----------
    eureqa_solution_id: str
        ID of this Solution
    complexity: int
        Complexity score for this solution. Complexity score is a function
        of the mathematical operators used in the current solution.
        The Complexity calculation can be tuned via model hyperparameters.
    error: float
        Error for the current solution, as computed by
        Eureqa using the 'error_metric' error metric.
    expression: str
        Eureqa model equation string.
    expression_annotated: str
        Eureqa model equation string with variable names tagged for easy identification.
    best_model: bool
        True, if the model is determined to be the best
    """
    _client = staticproperty(get_client)

    def __init__(
            self,
            eureqa_solution_id,
            complexity,
            error,
            expression,
            expression_annotated,
            best_model,
            project_id):
        self.eureqa_solution_id = eureqa_solution_id
        self.complexity = complexity
        self.error = error
        self.expression = expression
        self.expression_annotated = expression_annotated
        self.best_model = best_model
        self._project_id = project_id

    def create_model(self):
        """Add this solution to the leaderboard, if it is not already present."""
        url = 'projects/{}/eureqaModels/'.format(self._project_id)
        data = {'solutionId': self.eureqa_solution_id}
        self._client.post(url, data=data)


class ParetoFront(APIObject):
    """ Pareto front data for a Eureqa model.

    The pareto front reflects the tradeoffs between error and complexity for particular model. The
    solutions reflect possible Eureqa models that are different levels of complexity.  By default,
    only one solution will have a corresponding model, but models can be created for each solution.

    Attributes
    ----------
    project_id : str
        the ID of the project the model belongs to
    error_metric : str
        Eureqa error-metric identifier used to compute error metrics for this search. Note that
        Eureqa error metrics do NOT correspond 1:1 with DataRobot error metrics -- the available
        metrics are not the same, and are computed from a subset of the training data rather than
        from the validation data.
    hyperparameters : dict
        Hyperparameters used by this run of the Eureqa blueprint
    target_type : str
       Indicating what kind of modeling is being done in this project, either 'Regression',
       'Binary' (Binary classification), or 'Multiclass' (Multiclass classification).
    solutions : list(Solution)
        Solutions that Eureqa has found to model this data.
        Some solutions will have greater accuracy.  Others will have slightly
        less accuracy but will use simpler expressions.
    """

    _Solution = t.Dict({
        t.Key('eureqa_solution_id'): t.String,
        t.Key('complexity'): t.Int,
        t.Key('error'): t.Float,
        t.Key('expression'): t.String,
        t.Key('expression_annotated'): t.String,
        t.Key('best_model'): t.Bool
    }).ignore_extra('*')

    ParetoFrontWrapper = t.Dict({
        t.Key('project_id'): t.String,
        t.Key('error_metric'): t.String,
        t.Key('hyperparameters'): t.Dict,
        t.Key('target_type'): t.String,
        t.Key('solutions'): t.List(_Solution)
    }).ignore_extra('*')

    _converter = ParetoFrontWrapper

    def __init__(
            self,
            project_id,
            error_metric,
            hyperparameters,
            target_type,
            solutions):
        self.project_id = project_id
        self.error_metric = error_metric
        self.hyperparameters = hyperparameters
        self.target_type = target_type
        self.solutions = [
            Solution(project_id=project_id, **soln)
            for soln in solutions
        ]

    def __repr__(self):
        return 'ParetoFront({})'.format(self.project_id)



from six.moves.urllib.parse import quote
import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.utils.waiters import wait_for_async_resolution
from ..utils import deprecation_warning, encode_utf8_if_py2, from_api


class Feature(APIObject):
    """ A feature from a project's dataset

    These are features either included in the originally uploaded dataset or added to it via
    feature transformations.  In time series projects, these will be distinct from the
    :class:`ModelingFeature <datarobot.models.ModelingFeature>` s created during partitioning;
    otherwise, they will correspond to the same features.  For more information about input and
    modeling features, see the :ref:`time series documentation<input_vs_modeling>`.

    The ``min``, ``max``, ``mean``, ``median``, and ``std_dev`` attributes provide information about
    the distribution of the feature in the EDA sample data.  For non-numeric features or features
    created prior to these summary statistics becoming available, they will be None.  For features
    where the summary statistics are available, they will be in a format compatible with the data
    type, i.e. date type features will have their summary statistics expressed as ISO-8601
    formatted date strings.

    Attributes
    ----------
    id : int
        the id for the feature - note that `name` is used to reference the feature instead of `id`
    project_id : str
        the id of the project the feature belongs to
    name : str
        the name of the feature
    feature_type : str
        the type of the feature, e.g. 'Categorical', 'Text'
    importance : float or None
        numeric measure of the strength of relationship between the feature and target (independent
        of any model or other features); may be None for non-modeling features such as partition
        columns
    low_information : bool
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : int
        number of unique values
    na_count : int or None
        number of missing values
    date_format : str or None
        For Date features, the date format string for how this feature
        was interpreted, compatible with https://docs.python.org/2/library/time.html#time.strftime .
        For other feature types, None.
    min : str, int, float, or None
        The minimum value of the source data in the EDA sample
    max : str, int, float, or None
        The maximum value of the source data in the EDA sample
    mean : str, int, or, float
        The arithmetic mean of the source data in the EDA sample
    median : str, int, float, or None
        The median of the source data in the EDA sample
    std_dev : str, int, float, or None
        The standard deviation of the source data in the EDA sample
    time_series_eligible : bool
        Whether this feature can be used as the datetime partition column in a time series project.
    time_series_eligibility_reason : str
        Why the feature is ineligible for the datetime partition column in a time series project,
        or 'suitable' when it is eligible.
    time_step : int or None
        For time series eligible features, a positive integer determining the interval at which
        windows can be specified. If used as the datetime partition column on a time series
        project, the feature derivation and forecast windows must start and end at an integer
        multiple of this value. None for features that are not time series eligible.
    time_unit : str or None
        For time series eligible features, the time unit covered by a single time step, e.g. 'HOUR',
        or None for features that are not time series eligible.
    target_leakage : str
        Whether a feature is considered to have target leakage or not.  A value of
        'SKIPPED_DETECTION' indicates that target leakage detection was not run on the feature.
        'FALSE' indicates no leakage, 'MODERATE' indicates a moderate risk of target leakage, and
        'HIGH_RISK' indicates a high risk of target leakage

    """
    _converter = t.Dict({
        t.Key('id'): t.Int,
        t.Key('project_id'): t.String,
        t.Key('name'): t.String,
        t.Key('feature_type', optional=True): t.String,
        t.Key('importance', optional=True): t.Float,
        t.Key('low_information'): t.Bool,
        t.Key('unique_count'): t.Int,
        t.Key('na_count', optional=True): t.Int,
        t.Key('date_format', optional=True): t.String,
        t.Key('min', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('max', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('mean', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('median', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('std_dev', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('time_series_eligible'): t.Bool,
        t.Key('time_series_eligibility_reason'): t.String,
        t.Key('time_step', optional=True): t.Int,
        t.Key('time_unit', optional=True): t.String,
        t.Key('target_leakage', optional=True): t.String
    }).allow_extra('*')

    def __init__(self, id, project_id=None, name=None, feature_type=None, importance=None,
                 low_information=None, unique_count=None, na_count=None, date_format=None,
                 min=None, max=None, mean=None, median=None, std_dev=None,
                 time_series_eligible=None, time_series_eligibility_reason=None,
                 time_step=None, time_unit=None, target_leakage=None):
        if isinstance(id, dict):
            deprecation_warning('Feature instantiation from a dict',
                                'v2.3',
                                'v3.0',
                                message='Use Feature.from_data instead')
            self.__init__(**from_api(id))
        else:
            self.id = id
            self.project_id = project_id
            self.name = name
            self.feature_type = feature_type
            self.importance = importance
            self.low_information = low_information
            self.unique_count = unique_count
            self.na_count = na_count
            self.date_format = date_format
            self.min = min
            self.max = max
            self.mean = mean
            self.median = median
            self.std_dev = std_dev
            self.time_series_eligible = time_series_eligible
            self.time_series_eligibility_reason = time_series_eligibility_reason
            self.time_step = time_step
            self.time_unit = time_unit
            self.target_leakage = target_leakage

    def __repr__(self):
        return encode_utf8_if_py2(u'Feature({})'.format(self.name))

    @classmethod
    def get(cls, project_id, feature_name):
        """Retrieve a single feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve

        Returns
        -------
        feature : Feature
            The queried instance
        """
        path = cls._build_url(project_id, feature_name)
        return cls.from_location(path)

    @classmethod
    def _build_url(cls, project_id, feature_name):
        return 'projects/{}/features/{}/'.format(
            project_id,
            feature_name if isinstance(feature_name, int) else quote(feature_name.encode('utf-8'))
        )

    def get_multiseries_properties(self, multiseries_id_columns, max_wait=DEFAULT_MAX_WAIT):
        """ Retrieve time series properties for a potential multiseries datetime partition column

        Multiseries time series projects use multiseries id columns to model multiple distinct
        series within a single project.  This function returns the time series properties (time step
        and time unit) of this column if it were used as a datetime partition column with the
        specified multiseries id columns, running multiseries detection automatically if it had not
        previously been successfully ran.

        Parameters
        ----------
        multiseries_id_columns : list of str
            the name(s) of the multiseries id columns to use with this datetime partition column.
            Currently only one multiseries id column is supported.
        max_wait : int, optional
            if a multiseries detection task is run, the maximum amount of time to wait for it to
            complete before giving up

        Returns
        -------
        properties : dict
            A dict with three keys:

                - time_series_eligible : bool, whether the column can be used as a partition column
                - time_unit : str or null, the inferred time unit if used as a partition column
                - time_step : int or null, the inferred time step if used as a partition column
        """
        def _extract_properties():
            retrieve_url = '{}multiseriesProperties/'.format(self._build_url(self.project_id,
                                                                             self.name))
            response = self._client.get(retrieve_url)
            response_schema = t.Dict({
                t.Key('datetimePartitionColumn'): t.String(),
                t.Key('detectedMultiseriesIdColumns'): t.List(t.Dict({
                    t.Key('multiseriesIdColumns'): t.List(t.String()),
                    t.Key('timeUnit'): t.String(),
                    t.Key('timeStep'): t.Int()
                }).ignore_extra('*'))
            }).ignore_extra('*')
            response_data = response_schema.check(response.json())
            detected_columns = response_data['detectedMultiseriesIdColumns']

            detected = next((col_set for col_set in detected_columns
                             if set(col_set['multiseriesIdColumns']) == set(multiseries_id_columns)
                             ), None)
            if detected is not None:
                return True, {'time_series_eligible': True,
                              'time_unit': detected['timeUnit'],
                              'time_step': detected['timeStep']}
            return False, {'time_series_eligible': False, 'time_unit': None, 'time_step': None}

        detect_url = 'projects/{}/multiseriesProperties/'.format(self.project_id)
        was_detected, properties = _extract_properties()
        if was_detected:
            return properties

        detection_job = self._client.post(detect_url,
                                          data={'datetimePartitionColumn': self.name,
                                                'multiseriesIdColumns': multiseries_id_columns}
                                          ).headers['Location']
        wait_for_async_resolution(self._client, detection_job, max_wait=max_wait)
        return _extract_properties()[1]


class ModelingFeature(APIObject):
    """ A feature used for modeling

    In time series projects, a new set of modeling features is created after setting the
    partitioning options.  These features are automatically derived from those in the project's
    dataset and are the features used for modeling.  Modeling features are only accessible once
    the target and partitioning options have been set.  In projects that don't use time series
    modeling, once the target has been set, ModelingFeatures and Features will behave
    the same.

    For more information about input and modeling features, see the
    :ref:`time series documentation<input_vs_modeling>`.

    As with the :class:`Feature <datarobot.models.Feature>` object, the `min`, `max, `mean`,
    `median`, and `std_dev` attributes provide information about the distribution of the feature in
    the EDA sample data.  For non-numeric features, they will be None.  For features where the
    summary statistics are available, they will be in a format compatible with the data type, i.e.
    date type features will have their summary statistics expressed as ISO-8601 formatted date
    strings.

    Attributes
    ----------
    project_id : str
        the id of the project the feature belongs to
    name : str
        the name of the feature
    feature_type : str
        the type of the feature, e.g. 'Categorical', 'Text'
    importance : float or None
        numeric measure of the strength of relationship between the feature and target (independent
        of any model or other features); may be None for non-modeling features such as partition
        columns
    low_information : bool
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : int
        number of unique values
    na_count : int or None
        number of missing values
    date_format : str or None
        For Date features, the date format string for how this feature
        was interpreted, compatible with https://docs.python.org/2/library/time.html#time.strftime .
        For other feature types, None.
    min : str, int, float, or None
        The minimum value of the source data in the EDA sample
    max : str, int, float, or None
        The maximum value of the source data in the EDA sample
    mean : str, int, or, float
        The arithmetic mean of the source data in the EDA sample
    median : str, int, float, or None
        The median of the source data in the EDA sample
    std_dev : str, int, float, or None
        The standard deviation of the source data in the EDA sample
    parent_feature_names : list of str
        A list of the names of input features used to derive this modeling feature.  In cases where
        the input features and modeling features are the same, this will simply contain the
        feature's name.  Note that if a derived feature was used to create this modeling feature,
        the values here will not necessarily correspond to the features that must be supplied at
        prediction time.
    """
    _converter = t.Dict({
        t.Key('project_id'): t.String,
        t.Key('name'): t.String,
        t.Key('feature_type', optional=True): t.String,
        t.Key('importance', optional=True): t.Float,
        t.Key('low_information'): t.Bool,
        t.Key('unique_count'): t.Int,
        t.Key('na_count', optional=True): t.Int,
        t.Key('date_format', optional=True): t.String,
        t.Key('min', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('max', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('mean', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('median', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('std_dev', optional=True): t.String | t.Int | t.Float | t.Null,
        t.Key('parent_feature_names'): t.List(t.String)
    }).ignore_extra('*')

    def __init__(self, project_id=None, name=None, feature_type=None, importance=None,
                 low_information=None, unique_count=None, na_count=None, date_format=None, min=None,
                 max=None, mean=None, median=None, std_dev=None, parent_feature_names=None):
        self.project_id = project_id
        self.name = name
        self.feature_type = feature_type
        self.importance = importance
        self.low_information = low_information
        self.unique_count = unique_count
        self.na_count = na_count
        self.date_format = date_format
        self.min = min
        self.max = max
        self.mean = mean
        self.median = median
        self.std_dev = std_dev
        self.parent_feature_names = parent_feature_names

    def __repr__(self):
        return encode_utf8_if_py2(u'ModelingFeature({})'.format(self.name))

    @classmethod
    def get(cls, project_id, feature_name):
        """Retrieve a single modeling feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve

        Returns
        -------
        feature : ModelingFeature
            The requested feature
        """
        path = cls._build_url(project_id, feature_name)
        return cls.from_location(path)

    @classmethod
    def _build_url(cls, project_id, feature_name):
        return 'projects/{}/modelingFeatures/{}/'.format(
            project_id,
            quote(feature_name.encode('utf-8'))
        )
